import pytest
import panel as pn
from momics.panel_utils import diversity_select_widgets


@pytest.fixture
def sample_columns():
    """
    Fixture that provides sample categorical and numerical columns for testing.
    """
    cat_columns = ["factor1", "factor2", "factor3"]
    num_columns = ["num1", "num2", "num3"]
    return cat_columns, num_columns


def test_diversity_select_widgets(sample_columns):
    """
    Tests the diversity_select_widgets function.
    """
    cat_columns, num_columns = sample_columns
    widgets = diversity_select_widgets(cat_columns, num_columns)

    # Check if the result is a tuple
    assert isinstance(widgets, tuple), "The result should be a tuple"

    # Check if the tuple contains 5 elements
    assert len(widgets) == 6, "The tuple should contain 5 elements"

    # Check if each element in the tuple is a pn.widgets.Select instance
    for widget in widgets[:-1]:
        assert isinstance(
            widget, pn.widgets.Select
        ), f"Expected pn.widgets.Select, but got {type(widget)}"

    # Check if the last element in the tuple is a pn.widgets.Checkbox instance
    assert isinstance(
        widgets[-1], pn.widgets.Checkbox
    ), f"Expected pn.widgets.Checkbox, but got {type(widgets[-1])}"

    # Check if the widgets have the correct initial values and options
    assert (
        widgets[0].value == "go"
    ), "The initial value of the first widget should be 'go'"
    assert widgets[0].options == [
        "go",
        "go_slim",
        "ips",
        "ko",
        "pfam",
    ], "The options of the first widget are incorrect"

    assert (
        widgets[1].value == cat_columns[0]
    ), f"The initial value of the second widget should be '{cat_columns[0]}'"
    assert (
        widgets[1].options == cat_columns
    ), "The options of the second widget are incorrect"

    assert (
        widgets[2].value == "ssu"
    ), "The initial value of the third widget should be 'ssu'"
    assert widgets[2].options == [
        "ssu",
        "lsu",
    ], "The options of the third widget are incorrect"

    assert (
        widgets[3].value == "ncbi_tax_id"
    ), "The initial value of the fourth widget should be 'ncbi_tax_id'"
    assert widgets[3].options == [
        "ncbi_tax_id",
        "superkingdom",
        "kingdom",
        "phylum",
        "class",
        "order",
        "family",
        "genus",
        "species",
    ], "The options of the fourth widget are incorrect"

    full_columns = sorted(num_columns + cat_columns)
    assert (
        widgets[4].value == full_columns[0]
    ), f"The initial value of the fifth widget should be '{full_columns[0]}'"
    assert (
        widgets[4].options == full_columns
    ), "The options of the fifth widget are incorrect"



import os
import logging
from datetime import datetime
from typing import List, Tuple

from bioblend.galaxy import GalaxyInstance
from bioblend.galaxy.config import ConfigClient


class RemGalaxy:
    def __init__(self, url_var_name: str, api_key_var_name: str):
        """Initializes the BlueCloud instance.

        Args:
            url_var_name (str): The name of the environment variable containing the Galaxy URL.
            api_key_var_name (str): The name of the environment variable containing the Galaxy API key.

        Raises:
            TypeError: If the provided Galaxy URL or API key is invalid.
            Exception: For any other errors that occur during initialization.
        """
        self.logger = logging.getLogger("BCGalaxy")
        self.logger.setLevel(logging.DEBUG)
        # logging.debug("test")

        # url is a name of the variable saved in the .env file
        # the same for the api_key
        try:
            self.set_galaxy_env(url_var_name, api_key_var_name)
            self.cfg = ConfigClient(self.gi)
            self.logger.info(f"User: {self.cfg.whoami()}")
            self.logger.info(f"Galaxy version: {self.cfg.get_version()}")
        except TypeError:
            self.logger.error("Please provide valid Galaxy URL and API key!")
        except Exception as e:
            self.logger.error(f"Error: {e}")

    def set_galaxy_env(self, url_var_name: str, api_key_var_name: str) -> List:
        """Sets the Galaxy environment variables.

        Args:
            url_var_name (str): The name of the environment variable containing the Galaxy URL.
            api_key_var_name (str): The name of the environment variable containing the Galaxy API key.

        Returns:
            List: A list of environment variables.
        """
        self.url = os.getenv(url_var_name)
        self.api_key = os.getenv(api_key_var_name)
        self.gi = GalaxyInstance(self.url, key=self.api_key)

    def get_datasets_by_key(
        self, key: str, value: str | List[str]
    ) -> Tuple[List, List, List]:
        """Retrieves datasets by a specific key and value.

        Args:
            key (str): The key to filter datasets.
            value (str | List[str]): The value to filter datasets.

        Returns:
            Tuple[List[str], List[str], List[str]]:
                A tuple containing:
                - A list of datasets (tuples) that match the key and value.
                - A list of dataset names that match the key and value.
                - A list of dataset IDs that match the key and value.
        """

        if isinstance(value, list):
            lst_dict = [
                k
                for k in self.gi.datasets.get_datasets()
                if key in k and k[key] in value
            ]
        elif isinstance(value, str):
            lst_dict = [
                k
                for k in self.gi.datasets.get_datasets()
                if key in k and k[key] == value
            ]

        else:
            raise ValueError("Value must be a string or a list of strings.")

        self.ds = [(k["name"], k["id"]) for k in lst_dict]
        self.ds_names = [k[0] for k in self.ds]
        self.ds_ids = [k[1] for k in self.ds]

        return self.ds, self.ds_names, self.ds_ids

    def get_datasets(self, name: str = None):
        """Retrieves datasets with an optional name filter.

        Args:
            name (str, optional): The name to filter datasets. Defaults to None.

        Returns:
            List[str]: A list of dataset names.
        """
        self.dataset_lst = self.gi.show_matching_datasets(
            self.history_id, name_filter=name
        )
        self.ds_names = [k["name"] for k in self.dataset_lst]

    def get_histories(self):
        """Retrieves all histories.

        Returns:
            List: A list of histories.
        """
        self.histories = self.gi.histories.get_histories()
        return self.histories

    def clean_histories_for_display(self):
        """Cleans the histories for display.

        Returns:
            List: A list of cleaned histories.
        """
        self.histories_cleaned = [
            {
                "id": k["id"],
                "name": k["name"],
                "last_updated": k["update_time"],
            }
            for k in self.histories
            if k["deleted"] != True
        ]
        return self.histories_cleaned

    def set_tool(self, tool_id: str):
        """Sets the tool ID.

        Args:
            tool_id (str): The ID of the tool.
        """
        self.tool_id = tool_id
        self.logger.info(f"Tool Info: {self.gi.tools.show_tool(self.tool_id)}")

    def set_history(self, create: bool = True, hid: str = None, hname: str = None):
        """Sets the history.

        Args:
            create (bool, optional): Whether to create a new history. Defaults to True.
            hid (str, optional): The ID of the history. Defaults to None.
            hname (str, optional): The name of the history. Defaults to None.
        """
        # name is id or the name of the newly created history
        if create:
            if hname is None:
                hname = f"History created at {datetime.now()}"
            self.history_name = hname
            self.history_id = self.gi.histories.create_history(hname)["id"]
            self.logger.info(
                f"History Info: {self.gi.histories.show_history(self.history_id)}"
            )
        else:  # user wants to use the existing history
            self.history_id = hid
            self.history_name = self.gi.histories.show_history(self.history_id)["name"]

    def set_dataset(self, dataset_id: str):
        """Sets the dataset ID.

        Args:
            dataset_id (str): The ID of the dataset.
        """
        self.dataset_id = dataset_id
        self.logger.info(
            f"Dataset Info: {self.gi.datasets.show_dataset(self.dataset_id)}"
        )

    ########
    # Jobs #
    ########
    def show_job_status(self, job_id: str):
        """Shows the status of a job.

        Args:
            job_id (str): The ID of the job.
        """
        raise NotImplementedError

    ###########
    # Actions #
    ###########
    def upload_file(self, file_path: str):
        """Uploads a file.

        Args:
            file_path (str): The path to the file to upload.
        """
        upload = self.gi.tools.upload_file(file_path, self.history_id)
        self.set_dataset(upload["outputs"][0]["id"])



import pytest
import pandas as pd
import matplotlib.pyplot as plt
from momics.plotting import *
from momics.diversity import alpha_diversity_parametrized, beta_diversity_parametrized
from skbio.stats.ordination import pcoa
from momics.test.fixtures import *


def test_plot_pcoa_black_no_color(sample_pcoa_df):
    """
    Tests the plot_pcoa_black function without coloring.
    """
    fig = plot_pcoa_black(sample_pcoa_df)

    # Check if the result is a matplotlib Figure
    assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

    # Check if the plot contains the correct labels and title
    ax = fig.axes[0]
    assert ax.get_xlabel() == "PC1", "The x-axis label should be 'PC1'"
    assert ax.get_ylabel() == "PC2", "The y-axis label should be 'PC2'"
    assert ax.get_title() == "PCoA Plot", "The title should be 'PCoA Plot'"

    # Check if the scatter plot is created with black color
    scatter = ax.collections[0]
    assert scatter.get_facecolor()[0, :3].tolist() == [
        0,
        0,
        0,
    ], "The points should be black"


def test_plot_pcoa_black_with_color(sample_pcoa_df):
    """
    Tests the plot_pcoa_black function with coloring.
    """
    fig = plot_pcoa_black(sample_pcoa_df, color_by="color_by")

    # Check if the result is a matplotlib Figure
    assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

    # Check if the plot contains the correct labels and title
    ax = fig.axes[0]
    assert ax.get_xlabel() == "PC1", "The x-axis label should be 'PC1'"
    assert ax.get_ylabel() == "PC2", "The y-axis label should be 'PC2'"
    assert (
        ax.get_title() == "PCoA Plot with valid color_by values: (100.00%)"
    ), "Wrong title "

    # Check if the scatter plot is created with the correct colormap
    scatter = ax.collections[0]
    assert scatter.get_cmap().name == "RdYlGn", "The colormap should be 'RdYlGn'"


def test_mpl_alpha_diversity(sample_alpha_df):
    """
    Tests the mpl_alpha_diversity function.
    """
    factor = "factor"
    fig = mpl_alpha_diversity(sample_alpha_df, factor=factor)

    # Check if the result is a matplotlib Figure
    assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

    # Check if the plot contains the correct labels and title
    ax = fig.axes[0]
    assert ax.get_xlabel() == "Sample", "The x-axis label should be 'Sample'"
    assert (
        ax.get_ylabel() == "Shannon Index"
    ), "The y-axis label should be 'Shannon Index'"
    assert (
        ax.get_title() == f"Shannon Index Grouped by {factor}"
    ), f"The title should be 'Shannon Index Grouped by {factor}'"

    # TODO: failing, butperhaps not so important, or restructure the test
    # Check if the bar plot is created with the correct data
    # bars = ax.patches
    # expected_heights = sample_alpha_df["Shannon"].tolist()
    # actual_heights = [bar.get_height() for bar in bars]
    # assert (
    #     actual_heights == expected_heights
    # ), f"Expected bar heights {expected_heights}, but got {actual_heights}"

    # # Check if the hue is applied correctly
    # expected_hues = sample_alpha_df[factor].tolist()
    # actual_hues = [bar.get_facecolor() for bar in bars]
    # unique_hues = sns.color_palette("coolwarm", len(set(expected_hues)))
    # expected_colors = [
    #     unique_hues[sorted(set(expected_hues)).index(hue)] for hue in expected_hues
    # ]
    # assert all(
    #     [
    #         actual_hue[:3] == expected_color
    #         for actual_hue, expected_color in zip(actual_hues, expected_colors)
    #     ]
    # ), "The hues are not applied correctly"


def test_mpl_average_per_factor(sample_alpha_df):
    """
    Tests the mpl_average_per_factor function.
    """
    factor = "factor"
    fig = mpl_average_per_factor(sample_alpha_df, factor=factor)

    # Check if the result is a matplotlib Figure
    assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

    # Check if the plot contains the correct labels and title
    ax = fig.axes[0]
    assert ax.get_xlabel() == factor, f"The x-axis label should be '{factor}'"
    assert (
        ax.get_ylabel() == "Shannon Index"
    ), "The y-axis label should be 'Shannon Index'"
    assert (
        ax.get_title() == f"Average Shannon Index Grouped by {factor}"
    ), f"The title should be 'Average Shannon Index Grouped by {factor}'"

    # Check if the bar plot is created with the correct data
    bars = ax.patches
    expected_heights = sample_alpha_df.groupby(factor)["Shannon"].mean().tolist()
    actual_heights = [bar.get_height() for bar in bars]
    assert (
        actual_heights == expected_heights
    ), f"Expected bar heights {expected_heights}, but got {actual_heights}"


@pytest.mark.parametrize(
    "table_name, col_to_add, order, backend",
    [
        ("go", "id", "factor", "hvplot"),
        ("go_slim", "id", "factor", "hvplot"),
        ("ips", "accession", "values", "matplotlib"),
        ("ko", "entry", "values", "matplotlib"),
        ("pfam", "entry", "values", "matplotlib"),
    ],
)
def test_alpha_plot(table_name, col_to_add, order, backend, sample_metadata):
    """
    Tests the alpha_plot function.
    """
    factor = "factor1"
    data_dict = sample_tables_dict(table_name, add_abundance=True)
    data_dict = {
        table_name: add_column(data_dict[table_name], col_to_add),
    }
    fig_pane = alpha_plot(
        data_dict,
        table_name,
        factor,
        sample_metadata,
        order=order,
        backend=backend,
    )

    hash_pane = {"matplotlib": pn.pane.Matplotlib, "hvplot": pn.pane.HoloViews}
    # Check if the result is a panel Matplotlib pane
    assert isinstance(
        fig_pane, hash_pane[backend]
    ), f"The result should be a panel {backend} pane"

    # Extract the figure from the pane
    fig = fig_pane.object

    if backend == "hvplot":
        # Check if the result is a hvplot Figure
        assert isinstance(fig, hv.element.Bars), "The result should be a hvplot Figure"
    elif backend == "matplotlib":
        # Check if the result is a matplotlib Figure
        assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

        # Check if the plot contains the correct labels and title
        ax = fig.axes[0]
        assert ax.get_xlabel() == "Sample", "The x-axis label should be 'Sample'"
        assert (
            ax.get_ylabel() == "Shannon Index"
        ), "The y-axis label should be 'Shannon Index'"
        assert (
            ax.get_title() == f"Shannon Index Grouped by {factor}"
        ), f"The title should be 'Shannon Index Grouped by {factor}'"

    # TODO: do not understand, have to have better clearer data
    # Check if the bar plot is created with the correct data_dict
    # alpha_df = alpha_diversity_parametrized(
    #     data_dict, table_name, sample_metadata
    # )
    # bars = ax.patches
    # expected_heights = alpha_df["Shannon"].tolist()
    # actual_heights = [bar.get_height() for bar in bars]
    # assert (
    #     actual_heights == expected_heights
    # ), f"Expected bar heights {expected_heights}, but got {actual_heights}"

    # # Check if the hue is applied correctly
    # expected_hues = alpha_df[factor].tolist()
    # actual_hues = [bar.get_facecolor() for bar in bars]
    # unique_hues = sns.color_palette("coolwarm", len(set(expected_hues)))
    # expected_colors = [
    #     unique_hues[sorted(set(expected_hues)).index(hue)] for hue in expected_hues
    # ]
    # assert all(
    #     [
    #         actual_hue[:3] == expected_color
    #         for actual_hue, expected_color in zip(actual_hues, expected_colors)
    #     ]
    # ), "The hues are not applied correctly"


@pytest.mark.parametrize(
    "table_name, col_to_add, order, backend",
    [
        ("go", "id", "factor", "hvplot"),
        ("go_slim", "id", "factor", "hvplot"),
        ("ips", "accession", "values", "matplotlib"),
        ("ko", "entry", "values", "matplotlib"),
        ("pfam", "entry", "values", "matplotlib"),
    ],
)
def test_av_alpha_plot(table_name, col_to_add, order, backend, sample_metadata):
    """
    Tests the av_alpha_plot function.
    """
    factor = "factor1"
    data_dict = sample_tables_dict(table_name, add_abundance=True)
    data_dict = {
        table_name: add_column(data_dict[table_name], col_to_add),
    }
    fig_pane = av_alpha_plot(
        data_dict, table_name, factor, sample_metadata, order=order, backend=backend
    )

    hash_pane = {"matplotlib": pn.pane.Matplotlib, "hvplot": pn.pane.HoloViews}
    # Check if the result is a panel Matplotlib pane
    assert isinstance(
        fig_pane, hash_pane[backend]
    ), f"The result should be a panel {backend} pane"

    # Extract the figure from the pane
    fig = fig_pane.object

    if backend == "hvplot":
        # Check if the result is a hvplot Figure
        assert isinstance(fig, hv.element.Bars), "The result should be a hvplot Figure"

    elif backend == "matplotlib":
        # Check if the result is a matplotlib Figure
        assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

        # Check if the plot contains the correct labels and title
        ax = fig.axes[0]
        assert ax.get_xlabel() == factor, f"The x-axis label should be '{factor}'"
        assert (
            ax.get_ylabel() == "Shannon Index"
        ), "The y-axis label should be 'Shannon Index'"
        assert (
            ax.get_title() == f"Average Shannon Index Grouped by {factor}"
        ), f"The title should be 'Average Shannon Index Grouped by {factor}'"

    # TODO: learn the calculations and design simple enough example data
    # TODO: treat nans and zeros et.
    # Check if the bar plot is created with the correct data
    # alpha_df = alpha_diversity_parametrized(
    #     data_dict, table_name, sample_metadata
    # )
    # expected_heights = alpha_df.groupby(factor)["Shannon"].mean().tolist()
    # bars = ax.patches
    # actual_heights = [bar.get_height() for bar in bars]
    # assert (
    #     actual_heights == expected_heights
    # ), f"Expected bar heights {expected_heights}, but got {actual_heights}"

    # # Check if the hue is applied correctly
    # expected_hues = alpha_df[factor].unique().tolist()
    # actual_hues = [bar.get_facecolor() for bar in bars]
    # unique_hues = sns.color_palette("coolwarm", len(set(expected_hues)))
    # expected_colors = [
    #     unique_hues[sorted(set(expected_hues)).index(hue)] for hue in expected_hues
    # ]
    # assert all(
    #     [
    #         actual_hue[:3] == expected_color
    #         for actual_hue, expected_color in zip(actual_hues, expected_colors)
    #     ]
    # ), "The hues are not applied correctly"


@pytest.mark.parametrize(
    "norm, backend",
    [
        (True, "hvplot"),
        (False, "hvplot"),
        (True, "matplotlib"),
        (False, "matplotlib"),
    ],
)
def test_beta_plot(norm, backend):
    """
    Tests the beta_plot function.
    """
    table_name = "sample_table"
    taxon = "GO:0001"
    data = sample_tables_dict(table_name, add_abundance=True)
    data = {
        table_name: add_column(data[table_name], "ncbi_tax_id"),
    }
    fig_pane = beta_plot(data, table_name, norm, taxon, backend=backend)

    hash_pane = {"matplotlib": pn.pane.Matplotlib, "hvplot": pn.pane.HoloViews}
    # Check if the result is a panel Matplotlib pane
    assert isinstance(
        fig_pane, hash_pane[backend]
    ), f"The result should be a panel {backend} pane"

    # Extract the figure from the pane
    fig = fig_pane.object

    if backend == "hvplot":
        # Check if the result is a hvplot Figure
        assert isinstance(
            fig, hv.element.HeatMap
        ), "The result should be a hvplot Figure"
    elif backend == "matplotlib":
        # Check if the result is a matplotlib Figure
        assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

        # Check if the plot contains the correct labels and title
        ax = fig.axes[0]
        assert (
            ax.get_title() == f"Beta diversity for {taxon}"
        ), f"The title should be 'Beta diversity for {taxon}'"

        # Check if the heatmap is created with the correct data
        beta_df = beta_diversity_parametrized(
            data[table_name], taxon=taxon, metric="braycurtis"
        )
        heatmap_data = beta_df.to_data_frame()
        heatmap_values = ax.collections[0].get_array().reshape(heatmap_data.shape)
        expected_values = heatmap_data.values
        assert (
            heatmap_values == expected_values
        ).all(), "The heatmap values are not correct"


# TODO: too complicated, need to simplify
# def test_beta_plot_pc(sample_metadata):
#     """
#     Tests the beta_plot_pc function.
#     """
#     factor = "factor1"
#     table_name = "sample_table"
#     taxon = "GO:0001"
#     data = sample_tables_dict(table_name, add_abundance=True)
#     data = {
#         table_name: add_column(data[table_name], "ncbi_tax_id"),
#     }

#     fig_pane = beta_plot_pc(
#         data, sample_metadata, table_name, factor, taxon
#     )

#     # Check if the result is a panel Matplotlib pane
#     assert isinstance(
#         fig_pane, pn.pane.Matplotlib
#     ), "The result should be a panel Matplotlib pane"

#     # Extract the figure from the pane
#     fig = fig_pane.object

#     # Check if the result is a matplotlib Figure
#     assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

#     # Check if the plot contains the correct labels and title
#     ax = fig.axes[0]
#     assert ax.get_xlabel() == "PC1", "The x-axis label should be 'PC1'"
#     assert ax.get_ylabel() == "PC2", "The y-axis label should be 'PC2'"
#     assert ax.get_title() == "PCoA Plot", "The title should be 'PCoA Plot'"

#     # Check if the scatter plot is created with the correct data
#     beta_df = beta_diversity_parametrized(
#         data[table_name], taxon=taxon, metric="braycurtis"
#     )
#     pcoa_result = pcoa(beta_df, method="eigh", number_of_dimensions=3)
#     pcoa_df = pd.merge(
#         pcoa_result.samples,
#         sample_metadata,
#         left_index=True,
#         right_on="ref_code",
#         how="inner",
#     )
#     scatter = ax.collections[0]
#     assert all(
#         scatter.get_offsets()[:, 0] == pcoa_df["PC1"]
#     ), "The PC1 values are not correct"
#     assert all(
#         scatter.get_offsets()[:, 1] == pcoa_df["PC2"]
#     ), "The PC2 values are not correct"

#     # Check if the color is applied correctly
#     if factor:
#         assert (
#             scatter.get_array().tolist() == pcoa_df[factor].tolist()
#         ), "The color values are not correct"



import os
import pandas as pd
from typing import Dict
from mgo.udal import UDAL


def load_parquets(folder: str) -> Dict[str, pd.DataFrame]:
    """
    Loads all .parquet files in a folder and stores them in a dictionary.

    The keys of the dictionary are the file names without the .parquet
    extension. If the filename contains more than one '.', only the last part
    of the name is included.

    Example:
        metagoflow_analyses.go_slim.parquet -> key = "go_slim"

    Args:
        folder (str): The path to the folder containing the .parquet files.

    Returns:
        dict: A dictionary containing the data frames of the .parquet files.
    """
    # Create an empty dictionary to store the data frames
    # In this disctionary the data tables will be stored as pandas data frames
    mgf_parquet_dfs = {}

    # Loop through the folder and load each .parquet file
    for file_name in os.listdir(folder):
        if file_name.endswith(".parquet"):
            file_path = os.path.join(folder, file_name)
            # Load the parquet file into a DataFrame
            df = pd.read_parquet(file_path)
            # Use the file name without extension as the dictionary key
            name = file_name.split(".")[-2].lower()
            mgf_parquet_dfs[name] = df
    return mgf_parquet_dfs


def load_parquets_udal():
    """
    Load parquet files into a dictionary by looping udal calls
    """
    udal = UDAL()

    parquets = {}
    for dataset in ["go", "go_slim", "ips", "ko", "pfam", "lsu", "ssu"]:
        parquets[dataset] = udal.execute(f"urn:embrc.eu:emobon:{dataset}").data()
    return parquets



import pandas as pd
import numpy as np
import logging
from typing import List, Dict

import skbio
from skbio.diversity import beta_diversity

# from skbio.stats.ordination import pcoa
from skbio.stats.distance import permanova
from sklearn.metrics import pairwise_distances

logging.getLogger(__name__)


#########################
# Statistical functions #
#########################
def run_permanova(
    data: pd.DataFrame,
    metadata: pd.DataFrame,
    permanova_factor: str,
    permanova_group: List[str],
    permanova_additional_factors: List[str],
    permutations: int = 999,
    verbose: bool = False,
) -> Dict[str, pd.DataFrame]:
    """
    Run PERMANOVA on the given data and metadata.
    Args:
        data (pd.DataFrame): DataFrame containing the abundance data.
        metadata (pd.DataFrame): DataFrame containing the metadata.
        permanova_factor (str): The factor to use for PERMANOVA.
        permanova_group (List[str]): List of groups to include in the analysis.
        permanova_additional_factors (List[str]): Additional factors to test.
        permutations (int): Number of permutations for PERMANOVA. Default is 999.
        verbose (bool): If True, print detailed output.
    Returns:
        Dict[str, pd.DataFrame]: Dictionary containing PERMANOVA results for each factor.
    """
    # Filter metadata based on selected groups
    if permanova_factor == "All":
        filtered_metadata = metadata.copy()
    else:
        filtered_metadata = metadata[metadata[permanova_factor].isin(permanova_group)]

    if "ref_code" not in filtered_metadata.columns:
        raise KeyError("The 'ref_code' column is missing from filtered metadata.")

    # Match data and metadata samples
    matched_data = data[filtered_metadata["ref_code"]].T
    abundance_matrix = matched_data

    permanova_results = {}
    # factors_to_test = permanova_additional_factors
    for remaining_factor in permanova_additional_factors:
        factor_metadata = filtered_metadata.dropna(subset=[remaining_factor])
        combined_abundance = abundance_matrix.loc[factor_metadata["ref_code"]]

        # Calculate Bray-Curtis distance matrix
        dissimilarity_matrix = pairwise_distances(
            combined_abundance, metric="braycurtis"
        )
        distance_matrix_obj = skbio.DistanceMatrix(
            dissimilarity_matrix, ids=combined_abundance.index
        )

        factor_metadata = factor_metadata.set_index("ref_code")
        factor_metadata = factor_metadata.loc[
            factor_metadata.index.intersection(distance_matrix_obj.ids)
        ]

        if remaining_factor not in factor_metadata.columns:
            continue

        group_vector = factor_metadata[remaining_factor]
        if group_vector.nunique() < len(group_vector):
            if set(distance_matrix_obj.ids) == set(group_vector.index):
                permanova_result = permanova(
                    distance_matrix_obj,
                    grouping=group_vector,
                    permutations=permutations,
                )
                permanova_results[remaining_factor] = permanova_result
                if verbose:
                    print(f"Factor: {remaining_factor}")
                    logging.info(f"Factor: {remaining_factor}")
                    logging.info(
                        f"  F-statistic: {permanova_result['test statistic']:.4f}"
                    )
                    logging.info(f"  p-value: {permanova_result['p-value']:.4f}\n")
        else:
            logging.info(
                f"Skipping factor '{remaining_factor}' due to unique values in grouping vector."
            )

    return permanova_results


def shannon_index(row: pd.Series) -> float:
    """
    Calculates the Shannon index for a given row of data.

    Args:
        row (pd.Series): A row of data containing species abundances.

    Returns:
        float: The Shannon index value.
    """
    row = pd.to_numeric(row, errors="coerce")
    total_abundance = row.sum()
    if total_abundance == 0:
        return np.nan
    relative_abundance = row / total_abundance
    ln_relative_abundance = np.log(relative_abundance)
    ln_relative_abundance[relative_abundance == 0] = 0
    multi = relative_abundance * ln_relative_abundance * -1
    return multi.sum()  # Shannon entropy


def calculate_shannon_index(df: pd.DataFrame) -> pd.Series:
    """
    Applies the Shannon index calculation to each row of a DataFrame.

    Args:
        df (pd.DataFrame): A DataFrame containing species abundances.

    Returns:
        pd.Series: A Series containing the Shannon index for each row.
    """
    return df.apply(shannon_index, axis=1)


#######################
# diversity functions #
#######################
def calculate_alpha_diversity(df: pd.DataFrame, factors: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates the alpha diversity (Shannon index) for a DataFrame.

    Args:
        df (pd.DataFrame): A DataFrame containing species abundances.
        factors (pd.DataFrame): A DataFrame containing additional factors to merge.

    Returns:
        pd.DataFrame: A DataFrame containing the Shannon index and additional factors.
    """
    # Select columns that start with the appropriate prefix
    numeric_columns = [
        col
        for col in df.columns
        if col.startswith("GO:")
        or col.startswith("IPR")
        or col.startswith("K")
        or col.startswith("PF")
    ]

    # Calculate Shannon index only from the selected columns
    shannon_values = calculate_shannon_index(df[numeric_columns])

    # Create DataFrame with Shannon index and ref_code
    alpha_diversity_df = pd.DataFrame(
        {"ref_code": df["ref_code"], "Shannon": shannon_values}
    )

    # Merge with factors
    alpha_diversity_df = alpha_diversity_df.merge(factors, on="ref_code")

    return alpha_diversity_df


# alpha diversity
def alpha_diversity_parametrized(
    tables_dict: Dict[str, pd.DataFrame], table_name: str, metadata: pd.DataFrame
) -> pd.DataFrame:
    """
    Calculates the alpha diversity for a list of tables and merges with metadata.

    Args:
        tables_dict (Dict[str, pd.DataFrame]): A dictionary of DataFrames containing species abundances.
        table_name (str): The name of the table.
        metadata (pd.DataFrame): A DataFrame containing metadata.

    Returns:
        pd.DataFrame: A DataFrame containing the alpha diversity and metadata.
    """
    df_alpha_input = alpha_input(tables_dict, table_name).T.sort_values(by="ref_code")
    df_alpha_input = pd.merge(
        df_alpha_input, metadata, left_index=True, right_on="ref_code"
    )
    alpha = calculate_alpha_diversity(df_alpha_input, metadata)
    return alpha


def beta_diversity_parametrized(
    df: pd.DataFrame, taxon: str, metric: str = "braycurtis"
) -> pd.DataFrame:
    """
    Calculates the beta diversity for a DataFrame.

    Args:
        df (pd.DataFrame): A DataFrame containing species abundances.
        taxon (str): The taxon to use for the beta diversity calculation.
        metric (str, optional): The distance metric to use. Defaults to "braycurtis".

    Returns:
        pd.DataFrame: A DataFrame containing the beta diversity distances.
    """
    df_beta_input = diversity_input(df, kind="beta", taxon=taxon)
    beta = beta_diversity(metric, df_beta_input)
    return beta


####################
# helper functions #
####################
def update_subset_indicator(indicator, df):
    """Update the subset indicator with the number of unique ref_codes."""
    indicator.value = df["ref_code"].nunique()


def update_taxa_count_indicator(indicator, df):
    """Update the taxa count indicator with the number of unique taxa."""
    indicator.value = len(df)


# I think this is only useful for beta, not alpha diversity
def diversity_input(
    df: pd.DataFrame, kind: str = "alpha", taxon: str = "ncbi_tax_id"
) -> pd.DataFrame:
    """
    Prepare input for diversity analysis.

    Args:
        df (pd.DataFrame): The input dataframe.
        kind (str): The type of diversity analysis. Either 'alpha' or 'beta'.
        taxon (str): The column name containing the taxon IDs.

    Returns:
        pd.DataFrame: The input for diversity analysis.
    """
    # Convert DF
    out = pd.pivot_table(
        df,
        index="ref_code",
        columns=taxon,
        values="abundance",
        fill_value=0,
    )

    # Normalize rows
    if kind == "beta":
        out = out.div(out.sum(axis=1), axis=0)

    assert df.ncbi_tax_id.nunique(), out.shape[1]
    return out


# Function to get the appropriate column based on the selected table
# Valid table names: ['go', 'go_slim', 'ips', 'ko', 'pfam']
def get_key_column(table_name: str) -> str:
    """Returns the key column name based on the table name.

    Args:
        table_name (str): The name of the table.

    Returns:
        str: The key column name.

    Raises:
        ValueError: If the table name is unknown.
    """
    if table_name in ["go", "go_slim"]:
        return "id"
    elif table_name == "ips":
        return "accession"
    elif table_name in ["ko", "pfam"]:
        return "entry"
    else:
        raise ValueError(f"Unknown table: {table_name}")


def alpha_input(tables_dict: Dict[str, pd.DataFrame], table_name: str) -> pd.DataFrame:
    """
    Prepares the input data for alpha diversity calculation.

    Args:
        tables_dict (Dict[str, pd.DataFrame]): A dictionary of DataFrames containing species abundances.
        table_name (str): The name of the table to process.

    Returns:
        pd.DataFrame: A pivot table with species abundances indexed by the key column and ref_code as columns.
    """
    key_column = get_key_column(table_name)
    print("Key column:", key_column)

    # select distinct ref_codes from the dataframe
    ref_codes = tables_dict[table_name]["ref_code"].unique()
    print("length of the ref_codes:", len(ref_codes))
    out = pd.pivot_table(
        tables_dict[table_name],
        values="abundance",
        index=[key_column],
        columns=["ref_code"],
        aggfunc="sum",
        fill_value=0,
    )
    print("table shape:", out.shape)
    return out


# Example usage
# alpha_input = diversity_input(df, king='alpha')
# beta_input = diversity_input(df, king='beta')



import importlib as _importlib
import logging

# from momics.version import version as __version__

# This goes into your library somewhere
logging.getLogger("momics").addHandler(logging.NullHandler())

submoduless = [
    "diversity",
    "galaxy",
    "loader",
    "metadata",
    "plotting",
    "taxonomy",
    "networks",
]



"""
Methods to manipulate, concatenate, merge and enrich metadata files from EMO-BON samplings.

Some of these methods work as temporary solution to bad or incomplete data validation of the metadata tables.

Hopefully, that will not be the case for ever.
"""

import os
import pandas as pd
from typing import Dict, List
from datetime import datetime
from mgo.udal import UDAL


def get_metadata(folder):
    # Load metadata
    sample_metadata = pd.read_csv(
        os.path.join(folder, "Batch1and2_combined_logsheets_2024-11-12.csv")
    )

    observatory_metadata = pd.read_csv(
        os.path.join(folder, "Observatory_combined_logsheets_validated.csv")
    )

    # Merge metadata
    full_metadata = pd.merge(
        sample_metadata,
        observatory_metadata,
        on=["obs_id", "env_package"],  # Matching conditions
        how="inner",  # Inner join
    )

    # Sort the merged dataframe by 'ref_code' column in ascending order
    full_metadata = full_metadata.sort_values(by="ref_code", ascending=True)

    # first convert some of the boolean cols
    full_metadata["failure"] = full_metadata["failure"].astype(str)
    # replace the 'nan' values with 'NA'
    full_metadata["failure"] = full_metadata["failure"].replace("nan", "NA")

    # adding replacement for the missing values for object type columns
    full_metadata = fill_na_for_object_columns(full_metadata)

    return full_metadata


def get_metadata_udal():
    """
    Load metadata from the UDAL API
    """
    udal = UDAL()

    sample_metadata = udal.execute("urn:embrc.eu:emobon:logsheets").data()
    observatory_metadata = (
        udal.execute("urn:embrc.eu:emobon:observatories").data().set_index("obs_id")
    )

    # Merge metadata
    full_metadata = pd.merge(
        sample_metadata,
        observatory_metadata,
        on=["obs_id", "env_package"],  # Matching conditions
        how="inner",  # Inner join
    )

    # Sort the merged dataframe by 'ref_code' column in ascending order
    full_metadata = full_metadata.sort_values(by="ref_code", ascending=True)

    # first convert some of the boolean cols
    full_metadata["failure"] = full_metadata["failure"].astype(str)
    # replace the 'nan' values with 'NA'
    full_metadata["failure"] = full_metadata["failure"].replace("nan", "NA")

    # adding replacement for the missing values for object type columns
    full_metadata = fill_na_for_object_columns(full_metadata)

    return full_metadata


#####################
## Filter metadata ##
#####################
# Filter the metadata table based on the selections in box_granular
def filter_metadata_table(
    metadata_df: pd.DataFrame, selected_factors: Dict[str, List[str]]
) -> pd.DataFrame:
    """
    Filter the metadata DataFrame based on selected factors and their values.

    Args:
        metadata_df (pd.DataFrame): The metadata DataFrame to filter.
        selected_factors (Dict[str, List[str]]): A dictionary where keys are factor names and values are lists of selected values.
            If 'All' is in the list, that factor will not be filtered.
    Returns:
        pd.DataFrame: The filtered metadata DataFrame.
    """
    # Create a copy of the metadata DataFrame
    filtered_metadata = metadata_df.copy()
    # Apply filters for each selected factor
    for factor, selected_values in selected_factors.items():
        if "All" not in selected_values:
            filtered_metadata = filtered_metadata[
                filtered_metadata[factor].isin(selected_values)
            ]
    return filtered_metadata


## filter data according to the metadata
def filter_data(df: pd.DataFrame, filtered_metadata: pd.DataFrame) -> pd.DataFrame:
    """
    Filter the DataFrame based on the filtered metadata.
    This function filters the DataFrame columns based on the 'ref_code' values in the filtered metadata.

    Args:
        df (pd.DataFrame): The DataFrame to filter.
        filtered_metadata (pd.DataFrame): The filtered metadata DataFrame.
    Returns:
        pd.DataFrame: The filtered DataFrame.
    """
    # filter columns names of df which are in the filtered metadata
    cols_to_keep = list(
        [
            col
            for col in df.columns.str.strip()
            if col in filtered_metadata["ref_code"].to_list()
        ]
    )

    filtered_df = df[cols_to_keep]
    return filtered_df


######################
## Enhance metadata ##
######################
def enhance_metadata(
    metadata: pd.DataFrame, df_validation: pd.DataFrame
) -> pd.DataFrame:
    """
    Enhance the metadata DataFrame by processing the 'collection_date' column and extracting the season.
    This function also filters the metadata based on the 'ref_code' values in the df_validation DataFrame.

    Args:
        metadata (pd.DataFrame): The metadata DataFrame to enhance.
        df_validation (pd.DataFrame): The DataFrame containing valid samples for filtering.
    Returns:
        pd.DataFrame: The enhanced metadata DataFrame.
    """
    metadata = process_collection_date(metadata)
    metadata = extract_season(metadata)

    # Filter the metadata on the 'ref_code' only for entries that are in df_valid
    metadata = metadata[metadata["ref_code"].isin(df_validation["ref_code"])]

    missing = df_validation[~df_validation["ref_code"].isin(metadata["ref_code"])]
    assert len(missing) == 0, "Missing samples in the metadata"
    assert len(metadata) == len(
        df_validation
    ), "Filtered metadata does not match the valid samples"

    # add column to identify properly the replicates
    metadata["replicate_info"] = (
        metadata["obs_id"].astype(str)
        + "_"
        + metadata["env_package"].astype(str)
        + "_"
        + metadata["collection_date"].astype(str)
        + "_"
        + metadata["size_frac"].astype(str)
    )

    return metadata


def process_collection_date(metadata: pd.DataFrame) -> pd.DataFrame:
    """
    Process the 'collection_date' column in the metadata DataFrame.
    This function converts the 'collection_date' column to datetime format,
    extracts the year, month, and day, and adds them as new columns.
    It also converts the month number to the month name (abbreviated).

    Args:
        metadata (pd.DataFrame): The metadata DataFrame containing the 'collection_date' column.

    Returns:
        pd.DataFrame: The updated metadata DataFrame with new columns for year, month, and day.
    """
    # Convert the 'collection_date' column to datetime
    metadata["collection_date"] = metadata["collection_date"].apply(
        lambda x: datetime.strptime(x, "%Y-%m-%d") if x is not None else None
    )
    # Extract the year from the 'collection_date' column
    metadata["year"] = metadata["collection_date"].apply(
        lambda x: x.year if x is not None else None
    )
    # Extract the month from the 'collection_date' column
    metadata["month"] = metadata["collection_date"].apply(
        lambda x: x.month if x is not None else None
    )
    # Convert month to month name
    metadata["month_name"] = metadata["month"].apply(
        lambda x: (
            datetime.strptime(str(x), "%m").strftime("%B")[:3]
            if x is not None
            else None
        )
    )
    # Extract the day from the 'collection_date' column
    metadata["day"] = metadata["collection_date"].apply(
        lambda x: x.day if x is not None else None
    )
    return metadata


def extract_season(metadata: pd.DataFrame) -> pd.DataFrame:
    """
    Add a 'season' column to the metadata DataFrame.
    This function determines the season based on the 'month' and 'day' columns
    and adds it as a new column to the DataFrame.

    Args:
        metadata (pd.DataFrame): The metadata DataFrame containing 'month' and 'day' columns.

    Returns:
        pd.DataFrame: The updated metadata DataFrame with a new 'season' column.
    """
    # Extract the season based on the month and day
    metadata["season"] = metadata.apply(extract_season_single, axis=1)
    return metadata


def extract_season_single(row):
    """
    Determine the season based on the month and day.
    This function is used as a helper function for the apply method."
    """
    if (
        (row["month"] == 3 and row["day"] >= 21)
        or (row["month"] == 4)
        or (row["month"] == 5)
        or (row["month"] == 6 and row["day"] < 21)
    ):
        return "Spring"
    elif (
        (row["month"] == 6 and row["day"] >= 21)
        or (row["month"] == 7)
        or (row["month"] == 8)
        or (row["month"] == 9 and row["day"] < 23)
    ):
        return "Summer"
    elif (
        (row["month"] == 9 and row["day"] >= 23)
        or (row["month"] == 10)
        or (row["month"] == 11)
        or (row["month"] == 12 and row["day"] < 21)
    ):
        return "Autumn"
    else:  # Winter
        return "Winter"


def fill_na_for_object_columns(df):
    """
    Fill NA values with 'NA' for object columns in the dataframe.

    Args:
        df (pd.DataFrame): The input dataframe.

    Returns:
        pd.DataFrame: The dataframe with NA values filled for object columns.
    """
    # Apply fillna only to object columns
    df[df.select_dtypes(include=["object"]).columns] = df.select_dtypes(
        include=["object"]
    ).apply(lambda col: col.fillna("NA"))
    return df



import pytest
from momics.utils import *

"""
Testing utilities for resource monitoring, mainly memory usage and RAM load.
"""


# def test_memory_load(monkeypatch):
#     """
#     Tests the memory_load function.
#     """

#     class MockVirtualMemory:
#         def __init__(self, used, total):
#             self.used = used
#             self.total = total

#     def mock_virtual_memory():
#         return MockVirtualMemory(used=8 * (1024**3), total=16 * (1024**3))

#     monkeypatch.setattr("psutil.virtual_memory", mock_virtual_memory)

#     used_gb, total_gb = memory_load()

#     # Check if the used and total memory are correctly calculated
#     assert used_gb == 8, f"Expected used memory to be 8 GB, but got {used_gb} GB"
#     assert total_gb == 16, f"Expected total memory to be 16 GB, but got {total_gb} GB"


# def test_memory_usage(monkeypatch):
#     """
#     Tests the memory_usage function.
#     """

#     class MockVirtualMemory:
#         def __init__(self, percent):
#             self.percent = percent

#     def mock_virtual_memory():
#         return MockVirtualMemory(percent=50.0)

#     monkeypatch.setattr("psutil.virtual_memory", mock_virtual_memory)

#     percent_used = memory_usage()

#     # Check if the percentage of used memory is correctly calculated
#     assert (
#         percent_used == 50.0
#     ), f"Expected used memory percentage to be 50.0%, but got {percent_used}%"



import pytest
import pandas as pd
import numpy as np
from skbio.diversity import beta_diversity
import skbio
from momics.diversity import *
from momics.test.fixtures import *


@pytest.mark.parametrize("name", ["sample_table"])
def test_shannon_index(name):
    """Tests the shannon_index function."""
    data = sample_data(name)
    # Test with a row containing species abundances
    row = data.iloc[0]
    assert (
        row.all()
        == pd.Series(
            {
                "ref_code": "sample1",
                "GO:0001": 10,
                "GO:0002": 20,
                "IPR0001": 30,
                "K0001": 40,
                "PF0001": 50,
            }
        ).all()
    ), f"Expected {row}, but got {pd.Series(row)}"

    result = shannon_index(row[1:])
    expected = -sum(
        (x / row[1:].sum()) * np.log(x / row[1:].sum()) for x in row[1:] if x > 0
    )
    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"

    # Test with a row[1:] containing zero abundances
    row = data.iloc[1]
    result = shannon_index(row[1:])
    assert np.isnan(result), f"Expected NaN, but got {result}"

    # Test with a row[1:] containing equal abundances
    row = data.iloc[2]
    result = shannon_index(row[1:])
    expected = -sum(
        (x / row[1:].sum()) * np.log(x / row[1:].sum()) for x in row[1:] if x > 0
    )
    assert np.isclose(
        result, expected, equal_nan=True
    ), f"Expected {expected}, but got {result}"


def test_calculate_alpha_diversity(sample_factors):
    """Tests the calculate_alpha_diversity function."""
    data = sample_data()
    result = calculate_alpha_diversity(data, sample_factors)

    # Check if the result is a DataFrame
    assert isinstance(result, pd.DataFrame), "The result should be a DataFrame"

    # Check if the result contains the expected columns
    expected_columns = ["ref_code", "Shannon", "factor1"]
    assert all(
        col in result.columns for col in expected_columns
    ), f"Expected columns {expected_columns}, but got {result.columns.tolist()}"

    # Check if the Shannon index values are calculated correctly
    expected_shannon = data.apply(lambda row: shannon_index(row[1:]), axis=1)
    assert np.isclose(result["Shannon"], expected_shannon, equal_nan=True).all(), (
        "The Shannon_index values are not calculated correctly, diff is "
        + f"{(result['Shannon'] - expected_shannon).tolist()}"
    )

    # Check if the factors are merged correctly
    assert all(
        result["factor1"] == sample_factors["factor1"]
    ), "The factors are not merged correctly"


@pytest.mark.parametrize(
    "table_name, col_to_add",
    [
        ("go", "id"),
        ("go_slim", "id"),
        ("ips", "accession"),
        ("ko", "entry"),
        ("pfam", "entry"),
    ],
)
def test_alpha_diversity_parametrized(table_name, col_to_add, sample_factors):
    """Tests the alpha_diversity_parametrized function."""
    # data = sample_tables_dict(name)
    data_dict = sample_tables_dict(table_name, add_abundance=True)
    data_dict = {
        table_name: add_column(data_dict[table_name], col_to_add),
    }

    result = alpha_diversity_parametrized(data_dict, table_name, sample_factors)

    # Check if the result is a DataFrame
    assert isinstance(result, pd.DataFrame), "The result should be a DataFrame"

    # Check if the result contains the expected columns
    expected_columns = ["ref_code", "Shannon", "factor1"]
    assert all(
        col in result.columns for col in expected_columns
    ), f"Expected columns {expected_columns}, but got {result.columns.tolist()}"

    # Check if the Shannon index values are calculated correctly
    expected_shannon = data_dict[table_name].apply(
        lambda row: shannon_index(row[1:-1]), axis=1
    )

    # assert np.isclose(result["Shannon"], expected_shannon, equal_nan=True).all(), (
    #     "The Shannon_index values are not calculated correctly, diff is " +
    #     f"{(result['Shannon'] - expected_shannon).tolist()}"
    # )

    # Check if the factors are merged correctly
    assert all(
        result["factor1"] == sample_factors["factor1"]
    ), "The factors are not merged correctly"


def test_beta_diversity_parametrized():
    """
    Tests the beta_diversity_parametrized function.
    """
    taxon = "GO:0001"
    metric = "braycurtis"
    data = sample_data(add_abundance=True)

    # Mock the diversity_input function to return the input DataFrame
    def mock_diversity_input(df, kind, taxon):
        return df.set_index("ref_code")

    # Replace the diversity_input function with the mock
    from momics.diversity import diversity_input

    diversity_input = mock_diversity_input
    beta_input = diversity_input(data, kind="beta", taxon=taxon)
    expected_beta = beta_diversity(metric, beta_input)

    data["ncbi_tax_id"] = ["taxon1", "taxon2", "taxon3"]

    result = beta_diversity_parametrized(data, taxon, metric)

    # Check if the result is a DataFrame
    assert isinstance(
        result, skbio.stats.distance._base.DistanceMatrix
    ), "The result should be a DataFrame"

    # Check if the result contains the expected distances
    # TODO: does not work yet
    # pd.testing.assert_frame_equal(
    #     result.to_data_frame(), expected_beta.to_data_frame(), check_like=True
    # )


def test_diversity_input_alpha():
    """Tests the diversity_input function for alpha diversity."""
    data = sample_data(add_abundance=True)
    data["ncbi_tax_id"] = ["taxon1", "taxon2", "taxon3"]
    result = diversity_input(data, kind="alpha", taxon="ncbi_tax_id")

    # Check if the result is a DataFrame
    assert isinstance(result, pd.DataFrame), "The result should be a DataFrame"

    # Check if the result contains the expected shape
    expected_shape = (3, 3)  # 3 ref_codes and 3 taxons
    assert (
        result.shape == expected_shape
    ), f"Expected shape {expected_shape}, but got {result.shape}"

    # Check if the values are correct
    expected_values = {
        "taxon1": [50, 0, 0],
        "taxon2": [0, 0, 0],
        "taxon3": [0, 0, 5],
    }
    for taxon, values in expected_values.items():
        assert all(
            result[taxon] == values
        ), f"Expected values {values} for {taxon}, but got {result[taxon].tolist()}"


def test_diversity_input_beta():
    """
    Tests the diversity_input function for beta diversity.
    """
    data = sample_data(add_abundance=True)
    data["ncbi_tax_id"] = ["taxon1", "taxon2", "taxon3"]
    result = diversity_input(data, kind="beta", taxon="ncbi_tax_id")

    # Check if the result is a DataFrame
    assert isinstance(result, pd.DataFrame), "The result should be a DataFrame"

    # Check if the result contains the expected shape
    expected_shape = (3, 3)  # 3 ref_codes and 3 taxons
    assert (
        result.shape == expected_shape
    ), f"Expected shape {expected_shape}, but got {result.shape}"

    # Check if the values are normalized correctly
    expected_values = {
        "taxon1": [1.0, np.nan, 0.0],
        "taxon2": [0.0, 1.0, 0.0],
        "taxon3": [0.0, 0.0, 1.0],
    }
    for taxon, values in expected_values.items():
        np.isclose(result[taxon], values, equal_nan=True),
        f"Expected values {values} for {taxon}, but got {result[taxon].tolist()}"


def test_get_key_column():
    """
    Tests the get_key_column function.
    """
    # Test known table names
    assert get_key_column("go") == "id", "Expected 'id' for table 'go'"
    assert get_key_column("go_slim") == "id", "Expected 'id' for table 'go_slim'"
    assert get_key_column("ips") == "accession", "Expected 'accession' for table 'ips'"
    assert get_key_column("ko") == "entry", "Expected 'entry' for table 'ko'"
    assert get_key_column("pfam") == "entry", "Expected 'entry' for table 'pfam'"

    # Test unknown table name
    with pytest.raises(ValueError, match="Unknown table: unknown_table"):
        get_key_column("unknown_table")


def test_alpha_input():
    """
    Tests the alpha_input function.
    """
    table_name = "go"
    data_dict = sample_tables_dict(table_name, add_abundance=True)
    data_dict = {
        table_name: add_column(data_dict[table_name], "id"),
    }
    result = alpha_input(data_dict, table_name)

    # Check if the result is a DataFrame
    assert isinstance(result, pd.DataFrame), "The result should be a DataFrame"

    # Check if the result contains the expected shape
    expected_shape = (3, 3)  # 3 ids and 3 ref_codes
    assert (
        result.shape == expected_shape
    ), f"Expected shape {expected_shape}, but got {result.shape}"

    # TODO: Check if the values are correct
    expected_values = {
        "sample1": [50, 0, 0],
        "sample2": [0, 0, 0],
        "sample3": [0, 0, 5],
    }
    for ref_code, values in expected_values.items():
        assert all(
            result[ref_code] == values
        ), f"Expected values {values} for {ref_code}, but got {result[ref_code].tolist()}"



import os
import panel as pn
from typing import List, Tuple
from pyngrok import ngrok
import pandas as pd
from IPython import get_ipython

from .utils import memory_load


def serve_app(template, env, name="panel app"):
    port = 4040
    while is_port_in_use(port):
        print("Port 4040 is in use, trying another port")
        port += 1
    print(f"Using port {port}")
    server = pn.serve(
        {name: template},
        port=port,
        address="127.0.0.1",
        threaded=True,
        websocket_origin="*",
    )

    if "google.colab" in str(get_ipython()) or env == "vscode":
        # server=pn.serve({"": template}, port=4040, address="127.0.0.1", threaded=True, websocket_origin="*")
        os.system(f"curl http://localhost:{port}")

        # Terminate open tunnels if exist
        ngrok.kill()

        # Setting the authtoken, get yours from https://dashboard.ngrok.com/auth
        NGROK_AUTH_TOKEN = os.getenv("NGROK_TOKEN")
        ngrok.set_auth_token(NGROK_AUTH_TOKEN)

        # Open an HTTPs tunnel on port 4040 for http://localhost:4040
        # if env == "vscode":
        public_url = ngrok.connect(addr=str(port))
        # else:
        #     public_url = ngrok.connect(port=str(port))

        print("Tracking URL:", public_url)
    else:
        pass
        # after development finished, this could be changed to np.serve()
        # server = pn.serve({name: template}, port=4040, address="127.0.0.1", threaded=True, websocket_origin="*")
        # template.servable()
    return server


def close_server(server, env):
    server.stop()
    if "google.colab" in str(get_ipython()) or env == "vscode":
        ngrok.disconnect(server)
        ngrok.kill()

    return


def is_port_in_use(port: int) -> bool:
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


def diversity_select_widgets(cat_columns: List[str], num_columns: List[str]) -> Tuple[
    pn.widgets.Select,
    pn.widgets.Select,
    pn.widgets.Select,
    pn.widgets.Select,
    pn.widgets.Select,
    pn.widgets.Checkbox,
]:
    """Creates selection widgets for alpha and beta diversity analysis.

    Args:
        cat_columns (List[str]): A list of categorical column names for alpha diversity.
        num_columns (List[str]): A list of numerical column names for beta diversity.

    Returns:
        Tuple[pn.widgets.Select, pn.widgets.Select, pn.widgets.Select, pn.widgets.Select, pn.widgets.Select]:
        A tuple containing selection widgets for alpha and beta diversity analysis.
    """
    select_table_alpha = pn.widgets.Select(
        name="Source table alphas",
        value="go",
        options=["go", "go_slim", "ips", "ko", "pfam"],
        description="Select a table for alpha diversity analysis",
    )

    select_cat_factor = pn.widgets.Select(
        name="Factor alpha",
        value=cat_columns[0],
        options=cat_columns,
        description="Categorical columns to compare alpha diversities",
    )

    select_table_beta = pn.widgets.Select(
        name="Source table beta",
        value="ssu",
        options=["ssu", "lsu"],
        description="Select a table for beta diversity analysis",
    )

    boptions = [
        "ncbi_tax_id",
        "superkingdom",
        "kingdom",
        "phylum",
        "class",
        "order",
        "family",
        "genus",
        "species",
    ]
    select_taxon = pn.widgets.Select(
        name="Taxon",
        value=boptions[0],
        options=boptions,
        description="At which taxon level is beta diversity calculated",
    )

    full_columns = sorted(num_columns + cat_columns)
    select_factor_beta_all = pn.widgets.Select(
        name="Factor beta",
        value=full_columns[0],
        options=full_columns,
        description="Factor to visualize beta PCoA towards",
    )
    ## checkbox for beta matrix normalization
    checkbox_beta_norm = pn.widgets.Checkbox(
        name="Normalize beta matrix",
        value=False,
    )

    ret = (
        select_table_alpha,
        select_cat_factor,
        select_table_beta,
        select_taxon,
        select_factor_beta_all,
        checkbox_beta_norm,
    )
    return ret


#####################
# Create indicators #
#####################
def create_indicators_landing_page(
    df: pd.DataFrame,
) -> List[pn.indicators.Number]:
    """
    Generates a list of indicators for the landing page based on the provided aggregated DataFrame.

    Args:
        df (pd.DataFrame): A DataFrame containing aggregated data with 'COMPLETED' and 'queued' columns.

    Returns:
        List[pn.indicators.Number]: A list of Panel indicators displaying the number of sequenced samples.
    """
    list_indicators = []

    nom1 = int(df["COMPLETED"]["filters"].sum())
    den1 = int(df["queued"]["filters"].sum()) + nom1
    list_indicators.append(
        pn.indicators.Number(
            name="Water samples sequenced",
            value=nom1,
            format="{value}" + f"/{den1}",
            width=150,
            font_size="34px",
            title_size="14px",
        ),
    )

    nom2 = int(df["COMPLETED"]["sediment"].sum())
    den2 = int(df["queued"]["sediment"].sum()) + nom2
    list_indicators.append(
        pn.indicators.Number(
            name="Sediment samples sequenced",
            value=nom2,
            format="{value}" + f"/{den2}",
            width=150,
            font_size="34px",
            title_size="14px",
        ),
    )

    try:
        nom3 = int(df["COMPLETED"]["water"].sum())
    except KeyError:
        nom3 = 0
    den3 = int(df["queued"]["filters_blank"].sum()) + nom3
    list_indicators.append(
        pn.indicators.Number(
            name="Blank Waters sequenced",
            value=nom3,
            format="{value}" + f"/{den3}",
            width=150,
            font_size="34px",
            title_size="14px",
        ),
    )

    try:
        nom4 = int(df["COMPLETED"]["sediment_blank"].sum())
    except KeyError:
        nom4 = 0
    den4 = int(df["queued"]["sediment_blank"].sum()) + nom4
    list_indicators.append(
        pn.indicators.Number(
            name="Blanks Sediments sequenced",
            value=nom4,
            format="{value}" + f"/{den4}",
            width=150,
            font_size="34px",
            title_size="14px",
        ),
    )
    return list_indicators


def create_indicators_diversity() -> (
    Tuple[pn.indicators.Progress, pn.indicators.Number]
):
    """Creates indicators for RAM usage.

    Returns:
        pn.FlexBox: A FlexBox containing RAM usage indicators.
    """
    used_gb, total_gb = memory_load()
    progress_bar = pn.indicators.Progress(
        name="Ram usage", value=int(used_gb / total_gb * 100), width=200
    )
    usage = pn.indicators.Number(
        value=used_gb,
        name="RAM usage [GB]",
        format="{value:,.2f}",
    )
    return progress_bar, usage



import pandas as pd
import numpy as np
from typing import List, Dict
from statsmodels.stats.multitest import multipletests
from skbio.stats import subsample_counts
from skbio.diversity import beta_diversity


"""
Some functions inspired by Andrzej Tkacz, must be cleaned up and documented
"""


def pivot_taxonomic_data(
    df: pd.DataFrame, normalize: str = None, rarefy_depth: int = None
) -> pd.DataFrame:
    """
    Prepares the taxonomic data (LSU and SSU tables) for analysis. Apart from
    pivoting.

    Normalization of the pivot is optional. Methods include:
    - None: no normalization.
    - 'tss_sqrt': Total Sum Scaling and Square Root Transformation.
    - 'rarefy': rarefaction to a specified depth, if None, min of sample sums is used.

    TODO: refactor scaling to a new method and offer different options.

    Args:
        df (pd.DataFrame): The input DataFrame containing taxonomic information.
        normalize (str, optional): Normalization method.
            Options: None, 'tss_sqrt', 'rarefy'. Defaults to None.
        rarefy_depth (int, optional): Depth for rarefaction. If None, uses min sample sum.
            Defaults to None.

    Returns:
        pd.DataFrame: A pivot table with taxonomic data.
    """
    # Select relevant columns
    df["taxonomic_concat"] = (
        df["ncbi_tax_id"].astype(str) +
        f";sk_{df["superkingdom"].fillna("")}" +
        f";k_{df["kingdom"].fillna("")}" +
        f";p_{df["phylum"].fillna("")}" +
        f";c_{df["class"].fillna("")}" +
        f";o_{df["order"].fillna("")}" +
        f";f_{df["family"].fillna("")}" +
        f";g_{df["genus"].fillna("")}" +
        f";s_{df["species"].fillna("")}"
    )
    pivot_table = (
        df.pivot_table(
            index=["ncbi_tax_id", "taxonomic_concat"],
            columns="ref_code",
            values="abundance",
        )
        .fillna(0)
        .astype(int)
    )
    pivot_table = pivot_table.reset_index()
    # change inex name
    pivot_table.columns.name = None

    # normalize values
    if normalize == "tss_sqrt":
        # Total Sum Scaling and Square Root Transformation
        pivot_table.iloc[:, 2:] = pivot_table.iloc[:, 2:].apply(lambda x: x / x.sum())
        pivot_table.iloc[:, 2:] = pivot_table.iloc[:, 2:].apply(lambda x: np.sqrt(x))
    elif normalize == "rarefy":
        # rarefy
        pivot_table.iloc[:, 2:] = rarefy_table(
            pivot_table.iloc[:, 2:], depth=rarefy_depth
        )
    else:
        # no normalization
        pass
    return pivot_table


def separate_taxonomy(
    df: pd.DataFrame, eukaryota_keywords: List[str] = None
) -> Dict[str, pd.DataFrame]:
    """
    Separate the taxonomic data into different categories based on the index names.
    Args:
        df (pd.DataFrame): The input DataFrame containing taxonomic information (LSU/SSU tables).
        eukaryota_keywords (List[str]): List of keywords to filter Eukaryota data.
    Returns:
        Dict[str, pd.DataFrame]: A dictionary containing separate DataFrames for Prokaryotes and Eukaryota.
    """
    # Separate rows based on "Bacteria", "Archaea", and "Eukaryota" entries
    prokaryotes_all = df[df.index.str.contains("Bacteria|Archaea", regex=True)]
    eukaryota_all = df[df.index.str.contains("Eukaryota", regex=True)]

    # Further divide "Prokaryotes all" into "Bacteria" and "Archaea"
    bacteria = prokaryotes_all[prokaryotes_all.index.str.contains("Bacteria")]
    archaea = prokaryotes_all[prokaryotes_all.index.str.contains("Archaea")]

    # Apply taxonomy splitting to the index
    taxonomy_levels = bacteria.index.to_series().apply(split_taxonomy)
    taxonomy_df = pd.DataFrame(
        taxonomy_levels.tolist(),
        columns=["phylum", "class", "order", "family", "genus", "species"],
        index=bacteria.index,
    )

    # Combine taxonomy with the abundance data
    bacteria_data = pd.concat([taxonomy_df, bacteria], axis=1)

    # Aggregate at each taxonomic level and save to CSV
    taxonomic_levels = ["phylum", "class", "order", "family", "genus"]
    bacteria_levels_dict = {}
    for level in taxonomic_levels:
        aggregated_df = aggregate_by_taxonomic_level(bacteria_data, level)
        # Standardize the values so each column sums to 100
        aggregated_df_normalized = (
            aggregated_df.div(aggregated_df.sum(axis=0), axis=1) * 100
        )
        bacteria_levels_dict[f"Bacteria_{level}"] = aggregated_df_normalized

    all_data = {
        "Prokaryotes All": prokaryotes_all,
        "Eukaryota All": eukaryota_all,
        "Bacteria": bacteria,
        "Archaea": archaea,
    }
    # all_data.update(eukaryota_dict)
    all_data.update(bacteria_levels_dict)

    # If eukaryota keywords are provided, separate Eukaryota data
    if eukaryota_keywords:
        eukaryota_dict = separate_taxonomy_eukaryota(eukaryota_all, eukaryota_keywords)
        all_data.update(eukaryota_dict)

    return all_data


def separate_taxonomy_eukaryota(df: pd.DataFrame, eukaryota_keywords: List[str]):
    """
    Separate Eukaryota data into different files based on specific keywords.
    Args:
        df (pd.DataFrame): The input DataFrame containing taxonomic information (LSU/SSU tables).
        eukaryota_keywords (List[str]): List of keywords to filter Eukaryota data.

    Example keywords:
        eukaryota_keywords = ['Discoba', 'Stramenopiles', 'Rhizaria', 'Alveolata',
                              'Amorphea', 'Archaeoplastida', 'Excavata']
    """
    # Further divide "Eukaryota all" by specific keywords
    eukaryota_dict = {}
    for keyword in eukaryota_keywords:
        subset = df[df["taxonomic_concat"].str.contains(keyword)]
        eukaryota_dict[keyword] = subset

    return eukaryota_dict


def split_taxonomy(index_name: str) -> List[str]:
    """
    Splits the taxonomic string into its components and removes prefixes.

    Args:
        index_name (str): The taxonomic string to split.

    Returns:
        List[str]: A list of taxonomic levels.
    """
    # Remove anything before "Bacteria" or "Archaea"
    if "Bacteria" in index_name:
        taxonomy = index_name.split("Bacteria;", 1)[1].split(";")
    elif "Archaea" in index_name:
        taxonomy = index_name.split("Archaea;", 1)[1].split(";")
    else:
        taxonomy = []
    # Return a list with taxonomic levels up to species
    return taxonomy[1:7]  # ['phylum', 'class', 'order', 'family', 'genus', 'species']


def aggregate_by_taxonomic_level(df: pd.DataFrame, level: str) -> pd.DataFrame:
    """
    Aggregates the DataFrame by a specific taxonomic level and sums abundances across samples.

    Args:
        df (pd.DataFrame): The input DataFrame containing taxonomic information.
        level (str): The taxonomic level to aggregate by (e.g., 'phylum', 'class', etc.).

    Returns:
        pd.DataFrame: A DataFrame aggregated by the specified taxonomic level.
    """
    # Drop rows where the level is missing
    df_level = df.dropna(subset=[level])
    # Group by the specified level and sum abundances across samples (columns)
    df_grouped = df_level.groupby(level).sum(numeric_only=True)
    return df_grouped


def remove_high_taxa(df: pd.DataFrame, tax_level: str = 'phylum') -> pd.DataFrame:
    """
    Remove high level taxa from the dataframe.

    Args:
        df (pd.DataFrame): DataFrame containing taxonomic data.
        tax_level (str): The taxonomic level to filter by (e.g., 'phylum', 'class', 'order', etc.).

    Returns:
        pd.DataFrame: DataFrame with rows where the specified taxonomic level is not None.
    """
    if tax_level not in df.columns:
        raise ValueError(f"Taxonomic level '{tax_level}' not found in DataFrame.")
    
    # Filter out rows where the taxonomic level is None or NaN
    return df[~df[tax_level].isna()].copy()


def prevalence_cutoff(
    df: pd.DataFrame, percent: float = 10, skip_columns: int = 2
) -> pd.DataFrame:
    """
    Apply a prevalence cutoff to the DataFrame, removing features that do not
    appear in at least a certain percentage of samples.
    This is useful for filtering out low-prevalence features that may not be
    biologically relevant.

    Args:
        df (pd.DataFrame): The input DataFrame containing feature abundances.
        percent (float): The prevalence threshold as a percentage.
        skip_columns (int): The number of columns to skip (e.g., taxonomic information).

    Returns:
        pd.DataFrame: A filtered DataFrame with low-prevalence features removed.
    """
    # Calculate the number of samples
    num_samples = df.shape[1] - skip_columns
    # Calculate the prevalence threshold
    threshold = (percent / 100) * num_samples
    # Filter features based on prevalence
    filtered = df.loc[df.iloc[:, skip_columns:].gt(0).sum(axis=1) >= threshold]
    # Reset the index
    filtered = filtered.reset_index(drop=True)
    return filtered


def rarefy_table(df: pd.DataFrame, depth: int = None, axis: int = 1) -> pd.DataFrame:
    """
    Rarefy an abundance table to a given depth. If depth is None, uses the
    minimum sample sum across all samples.
    This function is a wrapper around the skbio.stats.subsample_counts function.

    Args:
        df: pd.DataFrame (rows: features, columns: samples)
        depth: int or None, rarefaction depth. If None, uses min sample sum.
        axis: int, 1 for samples in columns, 0 for samples in rows.

    Returns:
        pd.DataFrame: A rarefied DataFrame.
    """
    if axis == 1:
        sample_sums = df.sum(axis=0)
    else:
        sample_sums = df.sum(axis=1)
    if depth is None:
        depth = sample_sums.min()
    rarefied = {}
    print("Minimum rarefaction depth:", depth)
    for sample in df.columns if axis == 1 else df.index:
        counts = df[sample].values if axis == 1 else df.loc[sample].values
        if sample_sums[sample] < depth:
            # Not enough counts, fill with NaN or zeros
            rarefied[sample] = np.full_like(counts, np.nan)
        else:
            rarefied_counts = subsample_counts(counts.astype(int), int(depth))
            rarefied[sample] = rarefied_counts
    if axis == 1:
        return pd.DataFrame(rarefied, index=df.index)
    else:
        return pd.DataFrame(rarefied, columns=df.columns)


def split_metadata(metadata: pd.DataFrame, factor: str) -> Dict[str, list]:
    """
    Splits the metadata ref codes to dictionary of key being the factor value and
    value is a list of the ref codes.

    Args:
        metadata (pd.DataFrame): The input DataFrame containing metadata.
        factor (str): The column name to split the metadata by.
    Returns:
        Dict[str, list]: A dictionary with keys as unique values of the factor and
                         values as lists of ref codes.
    """
    if factor not in metadata.columns:
        raise ValueError(f"Factor '{factor}' not found in DataFrame columns.")
    # check if column is categorical
    if not isinstance(metadata[factor].dtype, pd.CategoricalDtype):
        raise ValueError(f"Column '{factor}' is not categorical (object dtype).")

    # for each unique value in the factor column, create a new table and append to a dictionary
    grouped_data = {}
    for value in metadata[factor].unique():
        # filter the dataframe
        filtered_df = metadata[metadata[factor] == value]
        # get the ref codes
        ref_codes = filtered_df["ref_code"].tolist()
        # add to the dictionary
        grouped_data[value] = ref_codes

    return grouped_data


def split_taxonomic_data(
    taxonomy: pd.DataFrame, groups: Dict[str, list]
) -> Dict[str, pd.DataFrame]:
    """
    Splits the taxonomic data into dictionary of DataFrames for each group.
    The split is based on the ref_code column, which needs to be present in the
    dataframes.

    Args:
        df (pd.DataFrame): The input DataFrame containing taxonomic information.
        groups (Dict[str, list]): A dictionary where keys are unique values of a factor
            which correspond to the groups to split by.

    Returns:
        Dict[str, pd.DataFrame]: A dictionary with keys as unique values of the factor and
            values as DataFrames containing taxonomic data for each group.
    """
    if not isinstance(groups, dict):
        raise ValueError("Groups must be a dictionary.")

    # for each unique value in the factor column, create a new table and append to a dictionary
    grouped_data = {}
    for value, ref_codes in groups.items():
        # filter the dataframe
        filtered_df = taxonomy[taxonomy["ref_code"].isin(ref_codes)]
        # add to the dictionary
        grouped_data[value] = filtered_df

    return grouped_data


def split_taxonomic_data_pivoted(
    taxonomy: pd.DataFrame, groups: Dict[str, list]
) -> Dict[str, pd.DataFrame]:
    """
    Splits the taxonomic data into dictionary of DataFrames for each group.
    The split is based on the column names which need to match between the taxonomy DataFrame
    and the groups lists. The DataFrame should have a 'ncbi_tax_id' and 'taxonomic_concat' which
    will serve as index of the resulting DataFrames.

    Args:
        taxonomy (pd.DataFrame): The input DataFrame containing taxonomic information.
        groups (Dict[str, list]): A dictionary where keys are unique values of a factor
            which correspond to the groups to split by.

    Returns:
        Dict[str, pd.DataFrame]: A dictionary with keys as unique values of the factor and
            values as DataFrames with separate columns for each taxonomic rank.
    """
    if not isinstance(groups, dict):
        raise ValueError("Groups must be a dictionary.")

    # for each unique value in the factor column, create a new table and append to a dictionary
    grouped_data = {}
    for value, ref_codes in groups.items():
        # filter the dataframe
        filtered_df = taxonomy[["ncbi_tax_id", "taxonomic_concat"] + ref_codes]
        # remove rows with all zeros and print how many rows were removed
        len_before = len(filtered_df)
        filtered_df = filtered_df[filtered_df[ref_codes].sum(axis=1) != 0]
        len_after = len(filtered_df)
        print(f"Removed {len_before - len_after} rows with all zeros for {value}.")
        # check if the dataframe is empty
        if filtered_df.empty:
            print(f"Warning: No data for {value} in the taxonomic data.")
            continue

        # add to the dictionary
        grouped_data[value] = filtered_df

    return grouped_data


def compute_bray_curtis(
    df: pd.DataFrame, skip_cols: int = 2, direction: str = "samples"
) -> pd.DataFrame:
    """
    Compute Bray-Curtis dissimilarity and return as a pandas DataFrame.
    This function computes the Bray-Curtis dissimilarity for samples in the DataFrame.

    Args:
        df (pd.DataFrame): The input DataFrame containing sample counts.
        skip_cols (int): Number of columns to skip (e.g., taxonomic information).
        direction (str): Direction of the dissimilarity calculation, 'samples' or 'taxa'.

    Returns:
        pd.DataFrame: A DataFrame containing the Bray-Curtis dissimilarity matrix.
    """
    if direction not in ["samples", "taxa"]:
        raise ValueError("Direction must be either 'samples' or 'taxa'.")

    if direction == "samples":
        # Use the sample IDs as the index
        ids = df.columns[skip_cols:].astype(str).tolist()
        result = beta_diversity(
            metric="braycurtis", counts=df.iloc[:, skip_cols:].T, ids=ids
        )
    elif direction == "taxa":
        ids = df["ncbi_tax_id"].astype(str).tolist()
        result = beta_diversity(
            metric="braycurtis", counts=df.iloc[:, skip_cols:], ids=ids
        )

    bray_curtis_df = pd.DataFrame(result.data, index=ids, columns=ids)
    return bray_curtis_df


def fdr_pvals(p_spearman_df: pd.DataFrame) -> pd.DataFrame:
    """
    Apply FDR correction to the p-values DataFrame using Benjamini/Hochberg (non-negative)
    method. This function extracts the upper triangle of the p-values DataFrame.

    Args:
        p_spearman_df (pd.DataFrame): DataFrame containing p-values.

    Returns:
        pd.DataFrame: DataFrame with FDR corrected p-values.
    """
    # Extract upper triangle p-values
    pval_array = (
        p_spearman_df.where(np.triu(np.ones(p_spearman_df.shape), k=1).astype(bool))
        .stack()
        .values
    )

    # Apply FDR correction
    _rejected, pvals_corrected, _, _ = multipletests(
        pval_array, alpha=0.05, method="fdr_bh"
    )

    # Map corrected p-values back to a DataFrame
    pvals_fdr = p_spearman_df.copy()
    pvals_fdr.values[np.triu_indices_from(p_spearman_df, k=1)] = pvals_corrected
    pvals_fdr.values[np.tril_indices_from(p_spearman_df, k=0)] = (
        np.nan
    )  # Optional: keep only upper triangle
    return pvals_fdr



import io
import pandas as pd
from typing import Dict, List


def bytes_to_df(data: bytes, sep: str = "\t") -> Dict[str, pd.DataFrame]:
    """
    Convert a dictionary of bytes to a dictionary of DataFrames.

    Args:
        data (Dict[str, bytes]): A dictionary where keys are filenames and values are byte strings.
        sep (str): The separator used in the data files. Default is tab ("\t").

    Returns:
        Dict[str, pd.DataFrame]: A dictionary where keys are filenames and values are DataFrames.
    """
    df = pd.read_csv(
        io.StringIO(
            str(
                data,
                "utf-8",
            )
        ),
        sep=sep,
    )
    return df



import time
import json
import panel as pn

from momics.galaxy.bio_blend import RemGalaxy


class Gecco:
    def __init__(self, params):
        self.exp = RemGalaxy("GALAXY_EARTH_URL", "GALAXY_EARTH_KEY")
        self.logger = self.exp.logger
        self.gecco_tool_id = "gecco"
        self.gecco_tool_name = "GECCO"
        self.gecco_tool_version = "0.1.0"
        self.gecco_tool_description = "GECCO is a tool for comparative genomics."
        self.select_history = params["select_history"]
        self.current_history_name = params["current_history_name"]
        self.current_history_id = params["current_history_id"]
        self.select_dataset = params["select_dataset"]
        self.current_file_name = params["current_file_name"]
        self.current_file_id = params["current_file_id"]
        self.file_source_checkbox = params["file_source_checkbox"]
        self.file_input = params["file_input"]
        self.mask = params["mask"]
        self.cds = params["cds"]
        self.threshold = params["threshold"]
        self.postproc = params["postproc"]
        self.antimash_sideload = params["antimash_sideload"]
        self.history_name = params["history_name"]
        self.logger.info("GECCO initialized.")
        self.debug = False

    def handle_login(self, clicks):
        pn.state.notifications.info(f"User logged in: {self.exp.cfg.whoami()}")
        self.logger.info(f"You have clicked me {clicks} times")
        self.handle_get_histories(clicks)

        # get the datasets right upon login
        self.handle_get_datasets(clicks)

    def handle_get_histories(self, clicks):
        self.exp.get_histories()

        # clean histories dict for display, remove deleted histories and select fust relevant fields
        clean_histories = self.exp.clean_histories_for_display()

        self.select_history.options = clean_histories
        self.select_history.value = clean_histories[0]
        self.logger.info(f"{len(clean_histories)} histories found.")

        # update the current history name
        self.handle_update_current_history_name(self.select_history.value)
        self.handle_update_current_history_id(self.select_history.value)

    def handle_get_datasets(self, clicks):
        # filtering seems too narrow or too broad, not sure what to do about it, but broad is better
        datasets, *_ = self.exp.get_datasets_by_key(
            "extension", ["fasta", "gbk", "embl", "genbank", "gb"]
        )
        self.logger.info(f"Datasets found: {datasets}")
        # fill the select_dataset widget
        self.select_dataset.options = datasets
        self.select_dataset.value = datasets[0]
        if self.debug:
            self.logger.info(
                f"Datasets selector options: {self.select_dataset.options}"
            )
            self.logger.info(
                f"Datasets selector value: {self.select_dataset.value}, {type(self.select_dataset.value)}"
            )
            self.logger.info(
                f"Datasets selector value[0]: {self.select_dataset.value[0]}"
            )
            self.logger.info(
                f"Datasets selector value[1]: {self.select_dataset.value[1]}"
            )
        # this is the id of the dataset
        self.handle_update_current_file_name(self.select_dataset.value)

        self.logger.info(f"{len(datasets)} datasets found.")

    def handle_update_current_file_name(self, value):
        self.current_file_name.value = value[0]
        self.current_file_id.value = value[1]

    def handle_update_current_history_name(self, value):
        self.current_history_name.value = value["name"]

    def handle_update_current_history_id(self, value):
        self.current_history_id.value = value["id"]

    def handle_update_current_file_name(self, value):
        self.current_file_name.value = value[0]
        self.current_file_id.value = value[1]

    def handle_create_history(self, clicks):
        # this supresses history creation upon startup
        if clicks == 0:
            return

        if self.history_name.value != "":
            self.exp.set_history(hname=self.history_name.value)
        else:
            self.exp.set_history()

        # and update the select widget
        self.handle_get_histories(clicks)

    def handle_upload_dataset(self, clicks):
        if self.file_source_checkbox.value:
            pn.state.notifications.warning("You selected Galaxy source and not upload.")
        else:
            upload_data = self.exp.gi.tools.upload_file(
                self.file_input.value[0],
                self.select_history.value["id"],
            )

            # this needs to wait until the upload is done
            while (
                self.exp.gi.datasets.show_dataset(upload_data["outputs"][0]["id"])[
                    "state"
                ]
                != "ok"
            ):
                time.sleep(1)

            self.logger.info(f"Upload data: {upload_data}")
            uploaded_dataset_id = upload_data["outputs"][0]["id"]
            self.current_file_name.value = upload_data["outputs"][0]["name"]
            self.current_file_id.value = upload_data["outputs"][0]["id"]
            pn.state.notifications.success(f"Dataset {uploaded_dataset_id} uploaded.")
            self.logger.info(f"Dataset {uploaded_dataset_id} uploaded.")

    def handle_submit_gecco(self, clicks):
        if clicks == 0:
            pn.state.notifications.warning("You need to log in first.")
            return

        if self.current_file_id.value == "":
            pn.state.notifications.warning("No dataset selected.")
            self.logger.warning("No dataset selected.")
            return

        if self.current_history_id.value == "":
            pn.state.notifications.warning("No history selected.")
            self.logger.warning("No history selected.")
            return

        # create a submission dictionary
        submission_inputs = {
            "input": {
                "id": self.current_file_id.value,
                "src": "hda",
            },
            "mask": self.mask.value,
            "cds": self.cds.value,
            "threshold": self.threshold.value,
            "postproc": self.postproc.value,
            "antimash_sideload": self.antimash_sideload.value,
        }

        # Run the GECCO tool
        tool_run = self.exp.gi.tools.run_tool(
            history_id=self.current_history_id.value,
            tool_id=self.gecco_tool_id,
            tool_inputs=submission_inputs,
        )

        # Get job ID to monitor
        job_id = tool_run["jobs"][0]["id"]
        self.logger.info(f"GECCO tool job submitted with job ID: {job_id}")
        pn.state.notifications.info(f"GECCO tool job submitted with job ID: {job_id}")

        # add history ID and gecco ID to the submission_inputs
        submission_inputs["history_id"] = self.current_history_id.value
        submission_inputs["gecco_id"] = self.gecco_tool_id

        # save the job_id and submission details to local json file, name is the job_id
        with open(f"{job_id}.json", "w") as f:
            json.dump(submission_inputs, f)
        self.logger.info(f"Submission details saved to {job_id}.json")



import pytest
import os

from unittest.mock import MagicMock, patch
from momics.galaxy.bio_blend import RemGalaxy


@pytest.fixture
def bbgalaxy():
    """
    Fixture that provides a Galaxy instance for testing.
    """
    with patch.dict(
        os.environ,
        {"GALAXY_URL": "http://fake-galaxy-instance", "GALAXY_API_KEY": "fake_api_key"},
    ):
        return RemGalaxy(url_var_name="GALAXY_URL", api_key_var_name="GALAXY_API_KEY")


def test_init(bbgalaxy):
    """
    Tests the initialization of the Galaxy instance.
    """
    assert (
        bbgalaxy.url == "http://fake-galaxy-instance"
    ), "The Galaxy URL should be set correctly"
    assert (
        bbgalaxy.api_key == "fake_api_key"
    ), "The Galaxy API key should be set correctly"


def test_get_histories(monkeypatch, bbgalaxy):
    """
    Tests the get_histories method.
    """
    mock_histories = [
        {"id": "1", "name": "History 1"},
        {"id": "2", "name": "History 2"},
    ]
    mock_get_histories = MagicMock(return_value=mock_histories)
    monkeypatch.setattr(bbgalaxy.gi.histories, "get_histories", mock_get_histories)

    histories = bbgalaxy.get_histories()

    assert histories == mock_histories, "The histories should be retrieved correctly"


# TODO: Too hard to fix now
# def test_create_history(monkeypatch, bcgalaxy):
#     """
#     Tests the set_history method for creating a new history.
#     """
#     mock_history = {"id": "3", "name": "New History"}
#     mock_create_history = MagicMock(return_value=mock_history)
#     monkeypatch.setattr(bcgalaxy.gi.histories, "create_history", mock_create_history)

#     bcgalaxy.set_history(create=True, hname="New History")

#     assert bcgalaxy.history_id == "3", "The new history ID should be set correctly"
#     assert (
#         bcgalaxy.history_name == "New History"
#     ), "The new history name should be set correctly"


def test_set_existing_history(monkeypatch, bbgalaxy):
    """
    Tests the set_history method for setting an existing history.
    """
    mock_history = {"id": "3", "name": "Existing History"}
    mock_show_history = MagicMock(return_value=mock_history)
    monkeypatch.setattr(bbgalaxy.gi.histories, "show_history", mock_show_history)

    bbgalaxy.set_history(create=False, hid="3")

    assert bbgalaxy.history_id == "3", "The existing history ID should be set correctly"
    assert (
        bbgalaxy.history_name == "Existing History"
    ), "The existing history name should be set correctly"


# def test_get_datasets_by_key(monkeypatch, bbgalaxy):
#     """
#     Tests the get_datasets_by_key method.
#     """
#     mock_datasets = [
#         {"name": "Dataset 1", "id": "value165464", "value": "val x"},
#         {"name": "Dataset 2", "id": "valueasdfasdf", "value": "val x"},
#     ]
#     mock_get_datasets = MagicMock(return_value=mock_datasets)
#     monkeypatch.setattr(bbgalaxy.gi.datasets, "get_datasets", mock_get_datasets)

#     datasets = bbgalaxy.get_datasets_by_key(key="id", value="value165464")

#     assert datasets == [("Dataset 1", "value165464")], "The datasets should be retrieved correctly by key and value"


# TODO: Too hard to fix now
# def test_upload_file(monkeypatch, bcgalaxy):
#     """
#     Tests the upload_file method.
#     """
#     mock_upload = {"outputs": [{"id": "dataset_id"}]}
#     mock_upload_file = MagicMock(return_value=mock_upload)
#     monkeypatch.setattr(bcgalaxy.gi.tools, "upload_file", mock_upload_file)

#     bcgalaxy.history_id = "history_id"
#     bcgalaxy.upload_file(file_path="fake_path")

#     assert (
#         bcgalaxy.dataset_id == "dataset_id"
#     ), "The dataset ID should be set correctly after upload"



import pandas as pd


def interaction_to_graph(
    df: pd.DataFrame, pos_cutoff: float = 0.8, neg_cutoff: float = -0.6
) -> tuple:
    """
    Create a network from the correlation matrix.
    Args:
        df (pd.DataFrame): The input DataFrame containing correlation values.
        pos_cutoff (float): Positive correlation cutoff.
        neg_cutoff (float): Negative correlation cutoff.
    Returns:
        nodes (list): List of node indices.
        edges_pos (list): List of positive edges.
        edges_neg (list): List of negative edges.
    """
    nodes, edges_pos, edges_neg = [], [], []
    count_pos, count_neg = 0, 0
    cols = df.columns.tolist()
    for i in range(df.shape[0]):
        nodes.append(cols[i])
        for j in range(i + 1, df.shape[1]):
            if df.iloc[i, j] > pos_cutoff:
                edges_pos.append((cols[i], cols[j]))
                count_pos += 1
            # print(f"Sample {i} and Sample {j} have a high correlation of {df.iloc[i, j]}")
            elif df.iloc[i, j] < neg_cutoff:
                edges_neg.append((cols[i], cols[j]))
                count_neg += 1
                # print(f"Sample {i} and Sample {j} have a high negative correlation of {df.iloc[i, j]}")
    print(f"Number of positive edges: {count_pos}")
    print(f"Number of negative edges: {count_neg}")
    return nodes, edges_pos, edges_neg


def interaction_to_graph_with_pvals(
    df: pd.DataFrame,
    pvals_df: pd.DataFrame,
    pos_cutoff: float = 0.8,
    neg_cutoff: float = -0.6,
    p_val_cutoff: float = 0.05,
) -> tuple:
    """
    Create a network from the correlation matrix and p-values.
    Args:
        df (pd.DataFrame): The input DataFrame containing correlation values.
        pvals_df (pd.DataFrame): The DataFrame containing p-values.
        pos_cutoff (float): Positive correlation cutoff.
        neg_cutoff (float): Negative correlation cutoff.
    Returns:
        nodes (list): List of node indices.
        edges_pos (list): List of positive edges with p-values.
        edges_neg (list): List of negative edges with p-values.
    """
    nodes, edges_pos, edges_neg = [], [], []
    count_pos, count_neg = 0, 0
    cols = df.columns.tolist()
    for i in range(df.shape[0]):
        nodes.append(cols[i])
        for j in range(i + 1, df.shape[1]):
            if df.iloc[i, j] > pos_cutoff and pvals_df.iloc[i, j] < p_val_cutoff:
                edges_pos.append((cols[i], cols[j]))
                count_pos += 1
            elif df.iloc[i, j] < neg_cutoff and pvals_df.iloc[i, j] < p_val_cutoff:
                edges_neg.append((cols[i], cols[j]))
                count_neg += 1
    print(f"Number of positive edges: {count_pos}")
    print(f"Number of negative edges: {count_neg}")
    return nodes, edges_pos, edges_neg



import pytest
import pandas as pd

# import fastparquet

from momics.loader.parquets import load_parquets


@pytest.fixture
def folder(tmp_path):
    # Create a temporary folder
    folder = tmp_path / "data"
    folder.mkdir()
    # Create a sample DataFrame and save it as a parquet file
    df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    df.to_parquet(folder / "test.parquet")
    return folder


def test_load_parquets(folder):
    # Load the parquet file
    data = load_parquets(folder)
    # Check that the data is loaded correctly
    assert "test" in data
    assert isinstance(data["test"], pd.DataFrame)
    assert isinstance(data, dict)
    assert data["test"].equals(pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]}))



import pandas as pd
import pytest
from momics.plotting import *


def sample_tables_dict(name="sample_table", add_abundance=False):
    """Fixture that provides a dictionary of sample tables for testing.
    These need to start with 'GO', 'IPR', 'K', 'PF', which are hardcoded in
    the calculate_alpha_diversity function for instance.
    TODO: that is probably not that great, should raise warning for other columns
    Put in a separate test?
    """
    data = {
        "ref_code": ["sample1", "sample2", "sample3"],
        "GO:0001": [10, 0, 5],
        "GO:0002": [20, 0, 5],
        "IPR0001": [30, 0, 5],
        "K0001": [40, 0, 5],
        "PF0001": [50, 0, 5],
    }
    if add_abundance:
        data["abundance"] = [50, 0, 5]
    return {name: pd.DataFrame(data)}


def sample_data(name="sample_table", add_abundance=False):
    return sample_tables_dict(name, add_abundance)[name]


def add_column(df, name):
    """Add a column to a DataFrame."""
    df[name] = [str(k) for k in range(len(df))]
    return df


@pytest.fixture
def sample_factors():
    """Fixture that provides sample factors for testing."""
    factors = {
        "ref_code": ["sample1", "sample2", "sample3"],
        "factor1": ["A", "B", "C"],
    }

    return pd.DataFrame(factors)


@pytest.fixture
def sample_metadata():
    """
    Fixture that provides sample metadata for testing.
    """
    metadata = {
        "ref_code": ["sample1", "sample2", "sample3"],
        "factor1": ["A", "B", "C"],
    }
    return pd.DataFrame(metadata)


# @pytest.fixture
# def sample_tables_dict():
#     """
#     Fixture that provides a dictionary of sample tables for testing.
#     """
#     data = {
#         "ref_code": ["sample1", "sample2", "sample3"],
#         "GO:0001": [10, 0, 5],
#         "GO:0002": [20, 0, 5],
#         "IPR0001": [30, 0, 5],
#         "K0001": [40, 0, 5],
#         "PF0001": [50, 0, 5],
#         "abundance": [1, 1, 1],  # Add the 'abundance' column
#     }
#     return {"sample_table": pd.DataFrame(data)}


@pytest.fixture
def sample_alpha_df():
    """Fixture that provides sample alpha diversity data for testing."""
    data = {
        "ref_code": ["sample1", "sample2", "sample3"],
        "Shannon": [1.0, 2.0, 3.0],
        "factor": ["A", "B", "A"],
    }
    return pd.DataFrame(data)


@pytest.fixture
def sample_pcoa_df():
    """
    Fixture that provides sample PCoA data for testing.
    """
    data = {
        "PC1": [0.1, 0.2, 0.3],
        "PC2": [0.4, 0.5, 0.6],
        "color_by": [1, 2, 3],
    }
    return pd.DataFrame(data)


@pytest.fixture
def sample_beta_df():
    """
    Fixture that provides sample beta diversity data for testing.
    """
    data = {
        "sample1": [0.0, 0.5, 0.8],
        "sample2": [0.5, 0.0, 0.3],
        "sample3": [0.8, 0.3, 0.0],
    }
    return pd.DataFrame(data, index=["sample1", "sample2", "sample3"])


def test_mpl_plot_heatmap(sample_beta_df):
    """
    Tests the mpl_plot_heatmap function.
    """
    taxon = "GO:0001"
    fig = mpl_plot_heatmap(sample_beta_df, taxon)

    # Check if the result is a matplotlib Figure
    assert isinstance(fig, plt.Figure), "The result should be a matplotlib Figure"

    # Check if the plot contains the correct title
    ax = fig.axes[0]
    assert (
        ax.get_title() == f"Beta diversity for {taxon}"
    ), f"The title should be 'Beta diversity for {taxon}'"

    # Check if the heatmap is created with the correct data
    heatmap_values = ax.collections[0].get_array().reshape(sample_beta_df.shape)
    expected_values = sample_beta_df.values
    assert (
        heatmap_values == expected_values
    ).all(), "The heatmap values are not correct"



from .parquets import load_parquets, load_parquets_udal
from .ro_crates import (
    get_rocrate_metadata_gh,
    get_rocrate_data,
    extract_data_by_name,
    extract_all_datafiles,
)
from .utils import bytes_to_df


__all__ = [
    "get_rocrate_metadata_gh",
    "get_rocrate_data",
    "extract_data_by_name",
    "extract_all_datafiles",
    "load_parquets",
    "load_parquets_udal",
    "bytes_to_df",
]



"""
This module contains functions for plotting alpha and beta diversity results.

Functions:
- hvplot_alpha_diversity: Creates a horizontal bar plot for alpha diversity using hvplot.
- plot_pcoa_black: Plots a PCoA plot with optional coloring.
- mpl_alpha_diversity: Plots the Shannon index grouped by a factor.
- mpl_average_per_factor: Plots the average Shannon index grouped by a factor.
- mpl_bgcs_violin: Creates a violin plot for BGC probabilities by type.
- alpha_plot: Creates an alpha diversity plot.
- av_alpha_plot: Creates an average alpha diversity plot.
- beta_plot: Creates a beta diversity heatmap plot.
- beta_plot_pc: Creates a beta diversity PCoA plot.
- beta_plot_pc_granular: Creates a beta diversity PCoA plot for granular data.
- mpl_plot_heatmap: Creates a heatmap plot for beta diversity.
- fold_legend_labels_from_series: Folds a list of labels to a maximum length from a Series.
- change_legend_labels: Changes the labels of a legend on a given matplotlib axis.
- cut_xaxis_labels: Changes the x-tick labels by cutting them short.
- get_sankey: Creates a Sankey diagram using Plotly.

Constants:
- PLOT_FACE_COLOR: The face color for the plot.

TODO: returns should be plt.figure and not pn.pane.Matplotlib, as already
implemented for beta_plot_pc() function.
"""

from typing import List, Tuple, Dict, Union
from textwrap import fill
import numpy as np
import panel as pn
import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt
import plotly.graph_objects as go
import holoviews as hv
import hvplot.pandas  # noqa

from bokeh.models import CategoricalColorMapper, ContinuousColorMapper
from bokeh.palettes import Category20, viridis

from skbio.stats.ordination import pcoa
from .diversity import (
    alpha_diversity_parametrized,
    beta_diversity_parametrized,
)

PLOT_FACE_COLOR = "#e6e6e6"
MARKER_SIZE = 16


##########
# HVplot #
##########
def hvplot_heatmap(
    df: pd.DataFrame, taxon: str, norm: bool = False
) -> hv.element.HeatMap:
    """
    Creates a heatmap plot for beta diversity using hvplot.

    Args:
        df (pd.DataFrame): DataFrame containing beta diversity distances.
        taxon (str): The taxon level for beta diversity calculation.
        norm (bool): Whether to normalize the data.

    Returns:
        hv.element.HeatMap: A heatmap plot of beta diversity.
    """
    # Create the heatmap using hvplot
    heatmap = df.hvplot.heatmap(
        cmap="viridis",
        colorbar=True,
        xlabel="Sample",
        ylabel="Sample",
        title=f"Beta Diversity ({taxon})",
    ).opts(
        xticks=0,  # remove xticks labels
        yticks=0,  # remove yticks labels
        show_legend=False,  # hide legend
    )
    if norm:
        heatmap.opts(clim=(0, 1.0))  # Set color limits for normalization
    else:
        heatmap.opts(
            clim=(df.min().min(), df.max().max())
        )  # Set color limits based on data
    return heatmap


def hvplot_alpha_diversity(alpha: pd.DataFrame, factor: str) -> hv.element.Bars:
    """
    Creates a horizontal bar plot for alpha diversity using hvplot.

    Args:
        alpha (pd.DataFrame): DataFrame containing alpha diversity data.
        factor (str): The column name to group by.

    Returns:
        hv.element.Bars: A horizontal bar plot of alpha diversity.
    """
    # Define the color mapper using Bokeh's CategoricalColorMapper
    if 2 < len(alpha[factor].unique()) <= 20:
        pal = Category20[
            len(alpha[factor].unique())
        ]  # Use the correct number of colors
    else:
        pal = viridis(len(alpha[factor].unique()))

    color_mapper = CategoricalColorMapper(
        factors=alpha[factor]
        .unique()
        .tolist(),  # Unique categories in the factor column
        palette=pal,
    )

    # Create the horizontal bar plot using hvplot
    fig = alpha.hvplot.barh(
        x="ref_code",
        y="Shannon",
        xlabel="Sample",
        ylabel="Shannon Index",
        title=f"Alpha Diversity ({factor})",
        color=factor,  # Use the factor column for coloring
    ).opts(
        yticks=0,  # remove yticks labels
        xaxis="top",
        cmap=color_mapper.palette,  # Apply the color mapper's palette
        legend_position="top_right",  # Adjust legend position
        tools=["hover"],  # Add hover tool for interactivity
        backend_opts={"plot.toolbar.autohide": True},
    )
    return fig


def hvplot_average_per_factor(alpha: pd.DataFrame, factor: str) -> hv.element.Bars:
    """
    Creates a horizontal bar plot for alpha diversity using hvplot.

    Args:
        alpha (pd.DataFrame): DataFrame containing alpha diversity data.
        factor (str): The column name to group by.

    Returns:
        hv.element.Bars: A horizontal bar plot of alpha diversity.
    """
    # Define the color mapper using Bokeh's CategoricalColorMapper
    if len(alpha[factor].unique()) <= 20:
        pal = Category20[
            len(alpha[factor].unique())
        ]  # Use the correct number of colors
    else:
        pal = viridis(len(alpha[factor].unique()))

    color_mapper = CategoricalColorMapper(
        factors=alpha[factor]
        .unique()
        .tolist(),  # Unique categories in the factor column
        palette=pal,
    )

    # Create the horizontal bar plot using hvplot
    fig = alpha.hvplot.bar(
        x=factor,
        y="Shannon",
        xlabel=factor,
        ylabel="Shannon Index",
        title=f"Average Shannon Index Grouped by {factor}",
        color=factor,  # Use the factor column for coloring
    ).opts(
        cmap=color_mapper.palette,  # Apply the color mapper's palette
        show_legend=False,
        # legend_position="top_right",  # Adjust legend position
        tools=["hover"],  # Add hover tool for interactivity
        backend_opts={"plot.toolbar.autohide": True},
    )
    return fig


def hvplot_plot_pcoa_black(
    pcoa_df: pd.DataFrame, color_by: str = None
) -> hv.element.Scatter:
    """
    Plots a PCoA plot with optional coloring using hvplot.

    Args:
        pcoa_df (pd.DataFrame): A DataFrame containing PCoA results.
        color_by (str, optional): The column name to color the points by. Defaults to None.

    Returns:
        hv.element.Scatter: The PCoA plot.
    """
    perc = pcoa_df[color_by].count() / len(pcoa_df[color_by]) * 100

    if 2 < len(pcoa_df[color_by].unique()) <= 20:
        if pcoa_df[color_by].dtype == "object":
            pal = Category20[
                len(pcoa_df[color_by].unique())
            ]  # Use the correct number of colors
            color_mapper = CategoricalColorMapper(
                factors=pcoa_df[color_by]
                .unique()
                .tolist(),  # Unique categories in the factor column
                palette=pal,
            )
        else:
            color_mapper = ContinuousColorMapper(
                palette="Turbo256",
                low=pcoa_df[color_by].min(),
                high=pcoa_df[color_by].max(),
            )
    else:
        if pcoa_df[color_by].dtype == "object":
            pal = viridis(len(pcoa_df[color_by].unique()))
            color_mapper = CategoricalColorMapper(
                factors=pcoa_df[color_by]
                .unique()
                .tolist(),  # Unique categories in the factor column
                palette=pal,
            )
        else:
            color_mapper = ContinuousColorMapper(
                palette="Turbo256",
                low=pcoa_df[color_by].min(),
                high=pcoa_df[color_by].max(),
            )

    if pcoa_df[color_by].count() >= 0:
        # Create the scatter plot using hvplot
        fig = pcoa_df.hvplot.scatter(
            x="PC1",
            y="PC2",
            color=color_by,  # Use the factor column for coloring
        )

    else:
        fig = pcoa_df.hvplot.scatter(
            x="PC1",
            y="PC2",
            color="black",  # Use color_by or black for coloring
        )

    fig = fig.opts(
        cmap=color_mapper.palette,  # if color_mapper else viridis(1),  # Apply the color mapper's palette
        xlabel="PC1",
        ylabel="PC2",
        title=f"PCoA Plot with valid {color_by} values: ({perc:.2f}%)",
        size=MARKER_SIZE,
        fill_alpha=0.5,
        # legend_position="top_right",  # Adjust legend position
        show_legend=False,
        tools=["hover"],  # Add hover tool for interactivity
        backend_opts={"plot.toolbar.autohide": True},
    )

    return fig


##############
# Matplotlib #
##############
def plot_pcoa_black(pcoa_df: pd.DataFrame, color_by: str = None) -> plt.Figure:
    """
    Plots a PCoA plot with optional coloring.

    Args:
        pcoa_df (pd.DataFrame): A DataFrame containing PCoA results.
        color_by (str, optional): The column name to color the points by. Defaults to None.

    Returns:
        plt.Figure: The PCoA plot.
    """
    flag_massage = False
    fig = plt.figure(figsize=(10, 6), facecolor=(0, 0, 0, 0))
    fig.patch.set_facecolor(PLOT_FACE_COLOR)
    ax = fig.add_subplot(111)

    if color_by is not None:
        labels = fold_legend_labels_from_series(pcoa_df[color_by], 35)
        print(pcoa_df[color_by].dtype)
        # BETA created now only for numerical
        if pcoa_df[color_by].dtype == "object":
            print("categorical")
            sns.scatterplot(
                data=pcoa_df,
                x="PC1",
                y="PC2",
                hue=color_by,
                palette="coolwarm",
                edgecolor="black",
            )
            ax = change_legend_labels(ax, labels)
        else:
            print("numerical", pcoa_df[color_by].count(), len(pcoa_df[color_by]))
            flag_massage = True
            perc = pcoa_df[color_by].count() / len(pcoa_df[color_by]) * 100
            scatter = plt.scatter(
                pcoa_df["PC1"],
                pcoa_df["PC2"],
                c=pcoa_df[color_by],
                cmap="RdYlGn",
                edgecolor="k",
            )
            plt.colorbar(scatter, label=color_by)
    else:
        print("single color")
        plt.scatter(pcoa_df["PC1"], pcoa_df["PC2"], color="black")

    ax.set_xlabel("PC1")
    ax.set_ylabel("PC2")
    if flag_massage:
        ax.set_title(f"PCoA Plot with valid {color_by} values: ({perc:.2f}%)")
    else:
        ax.set_title("PCoA Plot")
    plt.tight_layout()
    plt.close(fig)
    return fig


def mpl_alpha_diversity(alpha_df: pd.DataFrame, factor: str = None) -> plt.Figure:
    """Plots the Shannon index grouped by a factor.

    Args:
        alpha_df (pd.DataFrame): A DataFrame containing alpha diversity results.
        factor (str, optional): The column name to group by. Defaults to None.

    Returns:
        plt.Figure: The Shannon index plot.
    """
    fig = plt.figure(figsize=(10, 6), facecolor=(0, 0, 0, 0))
    fig.patch.set_facecolor(PLOT_FACE_COLOR)
    labels = fold_legend_labels_from_series(alpha_df[factor], 35)

    ax = fig.add_subplot(111)
    sns.barplot(
        data=alpha_df,
        x="ref_code",
        y="Shannon",
        hue=factor,
        palette="coolwarm",
    )

    # check axes and find which is have legend
    ax = change_legend_labels(ax, labels)
    ax.tick_params(axis="x", which="major", labelsize=np.log(4e5 / len(alpha_df)))

    ax.set_title(f"Shannon Index Grouped by {factor}")
    ax.set_xlabel("Sample")
    ax.set_ylabel("Shannon Index")

    plt.xticks(rotation=90)
    plt.tight_layout()
    plt.close(fig)
    return fig


def mpl_average_per_factor(df: pd.DataFrame, factor: str = None) -> plt.Figure:
    """Plots the average Shannon index grouped by a factor.

    Args:
        df (pd.DataFrame): A DataFrame containing alpha diversity results.
        factor (str, optional): The column name to group by. Defaults to None.

    Returns:
        plt.Figure: The average Shannon index plot.
    """
    fig = plt.figure(figsize=(10, 6), facecolor=(0, 0, 0, 0))
    fig.patch.set_facecolor(PLOT_FACE_COLOR)
    ax = fig.add_subplot(111)

    sns.barplot(
        data=df,
        x=factor,
        y="Shannon",
        hue=factor,
        capsize=0.1,
        palette="coolwarm",
    )

    ax.set_title(f"Average Shannon Index Grouped by {factor}")
    ax.set_xlabel(factor)
    ax.set_ylabel("Shannon Index")
    ax = cut_xaxis_labels(ax, 15)

    plt.xticks(rotation=90)
    plt.tight_layout()
    plt.close(fig)
    return fig


## gecco analysis ##
####################
def mpl_bgcs_violin(df: pd.DataFrame, normalize: bool = False) -> plt.Figure:
    fig = plt.figure(figsize=(10, 6), facecolor=(0, 0, 0, 0))
    fig.patch.set_facecolor(PLOT_FACE_COLOR)
    ax = fig.add_subplot(111)

    df["type"] = df["type"].fillna("Unknown")

    sns.swarmplot(
        data=df, x="type", y="average_p", hue="max_p", size=7, palette="coolwarm", ax=ax
    )
    sns.violinplot(
        data=df,
        x="type",
        y="average_p",
        inner=None,
        fill=False,
        color="black",
        ax=ax,
    )
    if normalize:
        ax.set_ylim(-0.1, 1.1)

    ax.set_title(f"Probabilities of identified BGCs by type")
    ax.set_xlabel("BGC type")
    ax.set_ylabel("Average probability")
    ax = cut_xaxis_labels(ax, 15)

    plt.tight_layout()
    plt.close(fig)
    return fig


def hvplot_bgcs_violin(df: pd.DataFrame, normalize: bool = False) -> hv.Overlay:
    """
    Creates a violin plot for BGC probabilities by type using hvplot.

    Args:
        df (pd.DataFrame): A DataFrame containing BGC data with columns 'type', 'average_p', and 'max_p'.
        normalize (bool): Whether to normalize the y-axis to the range [0, 1].

    Returns:
        hv.Overlay: An overlay of swarm and violin plots.
    """
    df["type"] = df["type"].fillna("Unknown")

    # Create the swarm plot
    swarm = df.hvplot.scatter(
        x="type",
        y="average_p",
        c="max_p",
        size=MARKER_SIZE,
        cmap="coolwarm",
        colorbar=True,
        title="Probabilities of identified BGCs by type",
        xlabel="BGC type",
        ylabel="Average probability",
        tools=["hover"],
    )

    # Create the violin plot
    violin = df.hvplot.violin(
        # x="type",
        y="average_p",
        by="type",
    ).opts(
        violin_fill_color=PLOT_FACE_COLOR,
        violin_alpha=0.5,
        violin_line_width=0.2,
        violin_visible=True,
    )

    # Combine the swarm and violin plots
    combined = violin * swarm

    # Apply normalization if needed
    if normalize:
        combined = combined.opts(yaxis="log", ylim=(-0.1, 1.1))

    return combined


def plot_domain_abundance(
    filtered_domains: pd.Series, abundance_min: int
) -> hv.element.Bars:
    """
    Plot the histogram of the number of pfam domains from the feature table using hvplot.

    Args:
        filtered_domains (pd.Series): A Series containing domain names as the index and their abundances as values.
        abundance_min (int): The minimum abundance threshold for domains to be included in the plot.

    Returns:
        hv.element.Bars: A horizontal bar plot of domain abundances.
    """
    plot = filtered_domains.hvplot.bar(
        xlabel="Domain",
        ylabel="Abundance",
        title=f"Domain Abundance (at least {abundance_min})",
        height=600,
        width=1400,
        cmap="viridis",
        tools=["hover"],
    ).opts(
        xticks=0,  # Remove xtick labels if too many
        show_legend=False,
    )
    return plot


def plot_tsne(X_embedded: np.ndarray, kmeans) -> hv.element.Scatter:
    """
    Plot the t-SNE embedding of the clusters using hvplot.

    Args:
        X_embedded (np.ndarray): The t-SNE embedded coordinates.
        kmeans: The fitted KMeans object containing cluster labels.

    Returns:
        hv.element.Scatter: The t-SNE plot of domain clusters.
    """
    # Create a DataFrame for hvplot
    tsne_df = pd.DataFrame(
        {
            "t-SNE 1": X_embedded[:, 0],
            "t-SNE 2": X_embedded[:, 1],
            "Cluster": kmeans.labels_,
        }
    )

    # Create the scatter plot using hvplot
    plot = tsne_df.hvplot.scatter(
        x="t-SNE 1",
        y="t-SNE 2",
        c="Cluster",
        cmap="viridis",
        colorbar=True,
        size=MARKER_SIZE,
        title="t-SNE of Domain Clusters",
        xlabel="t-SNE 1",
        ylabel="t-SNE 2",
        tools=["hover"],
    )
    return plot


##################
# Plot for panel #
##################
# Alpha diversity
def alpha_plot(
    tables_dict: Dict[str, pd.DataFrame],
    table_name: str,
    factor: str,
    metadata: pd.DataFrame,
    order: str = "factor",  # or values
    backend: str = "hvplot",  # Options: "matplotlib" or "hvplot"
) -> Union[pn.pane.Matplotlib, pn.pane.HoloViews]:
    """
    Creates an alpha diversity plot.

    Args:
        tables_dict (Dict[str, pd.DataFrame]): A dictionary of DataFrames containing species abundances.
        table_name (str): The name of the table to process.
        factor (str): The column name to group by.
        metadata (pd.DataFrame): A DataFrame containing metadata.
        order (str): The order of sorting the data. Can be "factor" or "value".
        backend (str): The plotting backend to use. Can be "matplotlib" or "hvplot".

    Returns:
        Union[pn.pane.Matplotlib, pn.pane.HoloViews]: A pane containing the alpha diversity plot.
    """
    alpha = alpha_diversity_parametrized(tables_dict, table_name, metadata)
    hash_sort = {"factor": factor, "values": "Shannon"}
    alpha = alpha.sort_values(by=hash_sort[order])

    if backend == "matplotlib":
        fig = pn.pane.Matplotlib(
            mpl_alpha_diversity(alpha, factor=factor),
            sizing_mode="stretch_both",
            name="Alpha div",
        )
    elif backend == "hvplot":
        fig = pn.pane.HoloViews(
            hvplot_alpha_diversity(alpha, factor=factor),
            name="Alpha div",
            width=900,
            height=1500,
        )
    else:
        raise ValueError(f"Unknown backend: {backend}")
    return fig


def av_alpha_plot(
    tables_dict: Dict[str, pd.DataFrame],
    table_name: str,
    factor: str,
    metadata: pd.DataFrame,
    order: str = "factor",  # or values
    backend: str = "hvplot",  # Options: "matplotlib" or "hvplot"
) -> Union[pn.pane.Matplotlib, pn.pane.HoloViews]:
    """
    Creates an average alpha diversity plot.

    Args:
        tables_dict (Dict[str, pd.DataFrame]): A dictionary of DataFrames containing species abundances.
        table_name (str): The name of the table to process.
        factor (str): The column name to group by.
        metadata (pd.DataFrame): A DataFrame containing metadata.

    Returns:
        Union[pn.pane.Matplotlib, pn.pane.HoloViews]: A pane containing the average alpha diversity plot.
    """
    alpha = alpha_diversity_parametrized(tables_dict, table_name, metadata)
    # TODO: this will not work, because it gets grouped in mpl_average_per_factor I think
    hash_sort = {"factor": factor, "values": "Shannon"}
    alpha = alpha.sort_values(by=hash_sort[order])

    if backend == "matplotlib":
        fig = pn.pane.Matplotlib(
            mpl_average_per_factor(alpha, factor=factor),
            sizing_mode="stretch_both",
            name="AV Alpha div",
        )
    elif backend == "hvplot":
        fig = pn.pane.HoloViews(
            hvplot_average_per_factor(alpha, factor=factor),
            sizing_mode="stretch_both",
            name="AV Alpha div",
        )
    else:
        raise ValueError(f"Unknown backend: {backend}")
    return fig


def beta_plot(
    tables_dict: Dict[str, pd.DataFrame],
    table_name: str,
    norm: bool,
    taxon: str = "ncbi_tax_id",
    backend: str = "hvplot",  # Options: "matplotlib" or "hvplot"
) -> Union[pn.pane.Matplotlib, pn.pane.HoloViews]:
    """
    Creates a beta diversity heatmap plot.

    Args:
        tables_dict (Dict[str, pd.DataFrame]): A dictionary of DataFrames containing species abundances.
        table_name (str): The name of the table to process.
        taxon (str, optional): The taxon level for beta diversity calculation. Defaults to "ncbi_tax_id".
        norm (bool): Whether to normalize the data.
        backend (str): The plotting backend to use. Can be "matplotlib" or "hvplot".

    Returns:
        Union[pn.pane.Matplotlib, pn.pane.HoloViews]: A pane containing the beta diversity heatmap plot.
    """
    beta = beta_diversity_parametrized(
        tables_dict[table_name], taxon=taxon, metric="braycurtis"
    )

    if backend == "matplotlib":
        fig = pn.pane.Matplotlib(
            mpl_plot_heatmap(beta.to_data_frame(), taxon=taxon, norm=norm),
            sizing_mode="stretch_both",
            name="Beta div",
        )
    elif backend == "hvplot":
        fig = pn.pane.HoloViews(
            hvplot_heatmap(beta.to_data_frame(), taxon=taxon, norm=norm),
            sizing_mode="stretch_both",
            name="Beta div",
        )
    else:
        raise ValueError(f"Unknown backend: {backend}")
    return fig


def beta_plot_pc(
    tables_dict: Dict[str, pd.DataFrame],
    metadata: pd.DataFrame,
    table_name: str,
    factor: str,
    taxon: str = "ncbi_tax_id",
) -> Tuple[hv.element.Scatter, float]:
    """
    Creates a beta diversity PCoA plot.

    Args:
        tables_dict (Dict[str, pd.DataFrame]): A dictionary of DataFrames containing species abundances.
        metadata (pd.DataFrame): A DataFrame containing metadata.
        table_name (str): The name of the table to process.
        factor (str): The column name to color the points by.
        taxon (str, optional): The taxon level for beta diversity calculation. Defaults to "ncbi_tax_id".

    Returns:
        Tuple[hv.element.Scatter, float]: A tuple containing the beta diversity PCoA plot and the explained variance.
    """
    beta = beta_diversity_parametrized(
        tables_dict[table_name], taxon=taxon, metric="braycurtis"
    )
    pcoa_result = pcoa(beta, method="eigh")  # , number_of_dimensions=3)
    explained_variance = pcoa_result.proportion_explained[:2].sum() * 100
    pcoa_df = pd.merge(
        pcoa_result.samples,
        metadata,
        left_index=True,
        right_on="ref_code",
        how="inner",
    )
    return hvplot_plot_pcoa_black(pcoa_df, color_by=factor), explained_variance


def beta_plot_pc_granular(
    filtered_data: pd.DataFrame,
    metadata: pd.DataFrame,
    factor: str,
    # taxon: str = "ncbi_tax_id",
) -> Tuple[hv.element.Scatter, float]:
    """
    Creates a beta diversity PCoA plot.

    Args:
        tables_dict (Dict[str, pd.DataFrame]): A dictionary of DataFrames containing species abundances.
        metadata (pd.DataFrame): A DataFrame containing metadata.
        table_name (str): The name of the table to process.
        factor (str): The column name to color the points by.
        taxon (str, optional): The taxon level for beta diversity calculation. Defaults to "ncbi_tax_id".

    Returns:
        Tuple[plt.figure, float]: A tuple containing the beta diversity PCoA plot and the explained variance.
    """
    from skbio.diversity import beta_diversity

    beta = beta_diversity("braycurtis", filtered_data.iloc[:, 1:].T)
    pcoa_result = pcoa(beta, method="eigh")  # , number_of_dimensions=3)
    explained_variance = pcoa_result.proportion_explained[:2].sum() * 100
    pcoa_df = pd.merge(
        pcoa_result.samples,
        metadata,
        left_index=True,
        right_on="ref_code",
        how="inner",
    )

    # return plot_pcoa_black(pcoa_df, color_by=factor), explained_variance
    return hvplot_plot_pcoa_black(pcoa_df, color_by=factor), explained_variance


def mpl_plot_heatmap(df: pd.DataFrame, taxon: str, norm=False) -> plt.Figure:
    """
    Creates a heatmap plot for beta diversity.

    Args:
        df (pd.DataFrame): A DataFrame containing beta diversity distances.
        taxon (str): The taxon level for beta diversity calculation.
        norm (bool): Whether to normalize the data.

    Returns:
        plt.Figure: The heatmap plot.
    """
    fig = plt.figure(figsize=(10, 6), facecolor=(0, 0, 0, 0))
    fig.patch.set_facecolor(PLOT_FACE_COLOR)
    _ = fig.add_subplot(111)
    if norm:
        sns.heatmap(df, vmin=0, vmax=1.0, cmap="viridis")
    else:
        sns.heatmap(df, cmap="viridis")
    plt.title(f"Beta diversity for {taxon}")
    plt.tight_layout()
    plt.close(fig)
    return fig


####################
# Helper functions #
####################
def fold_legend_labels_from_series(df: pd.Series, max_len: int = 30) -> List[str]:
    """Folds a list of labels to a maximum length from a Series.

    Args:
        df (pd.Series): The series to extract unique labels.
        max_len (int, optional): The maximum length of a label. Defaults to 30.

    Returns:
        List[str]: The folded list of labels.
    """
    return [
        fill(x, max_len) if isinstance(x, str) and len(x) > max_len else str(x)
        for x in df.unique()
    ]


def change_legend_labels(ax: plt.axis, labels: List[str]) -> plt.axis:
    """Changes the labels of a legend on a given matplotlib axis.

        ax (plt.axis): The matplotlib axis object whose legend labels need to be changed.
        labels (List[str]): A list of new labels to be set for the legend.

    Returns:
        plt.axis: The matplotlib axis object with updated legend labels.
    """
    leg = ax.get_legend()
    for t, l in zip(leg.texts, labels):
        t.set_text(l)
    return ax


def cut_xaxis_labels(ax: plt.axis, n: int = 15) -> plt.axis:
    """Changes the x-tick labels by cutting them short.

    Args:
        ax: The axes to change the x-axis of.
        n: cutoff for max number of characters for the xtick label.

    Returns:
        plt.axis: The axes with the new x-tick labels.
    """
    ticks = ax.get_xticklabels()
    new_ticks = []
    for tick in ticks:
        tick.set_text(tick.get_text()[: min(len(tick.get_text()), n)])
        new_ticks.append(tick)
    ax.set_xticklabels(new_ticks)
    return ax


############
## Plotly ##
############
def get_sankey(df, cat_cols=[], value_cols="", title="Sankey Diagram"):
    # Colors
    colorPalette = [
        "rgba(31, 119, 180, 0.8)",
        "rgba(255, 127, 14, 0.8)",
        "rgba(44, 160, 44, 0.8)",
        "rgba(214, 39, 40, 0.8)",
        "rgba(148, 103, 189, 0.8)",
        "rgba(140, 86, 75, 0.8)",
        "rgba(227, 119, 194, 0.8)",
        "rgba(127, 127, 127, 0.8)",
    ]
    labelList = []
    colorNumList = []

    for catCol in cat_cols:
        labelListTemp = list(set(df[catCol].values))
        colorNumList.append(len(labelListTemp))
        labelList = labelList + labelListTemp

    # remove duplicates from labelList
    labelList = list(dict.fromkeys(labelList))

    # define colors based on number of levels
    colorList = []
    for idx, colorNum in enumerate(colorNumList):
        colorList = colorList + [colorPalette[idx]] * colorNum

    # transform df into a source-target pair
    for i in range(len(cat_cols) - 1):
        if i == 0:
            sourceTargetDf = df[[cat_cols[i], cat_cols[i + 1], value_cols]]
            sourceTargetDf.columns = ["source", "target", "count"]
        else:
            tempDf = df[[cat_cols[i], cat_cols[i + 1], value_cols]]
            tempDf.columns = ["source", "target", "count"]
            sourceTargetDf = pd.concat([sourceTargetDf, tempDf])
        sourceTargetDf = (
            sourceTargetDf.groupby(["source", "target"])
            .agg({"count": "sum"})
            .reset_index()
        )

    # add index for source-target pair
    sourceTargetDf["sourceID"] = sourceTargetDf["source"].apply(
        lambda x: labelList.index(x)
    )
    sourceTargetDf["targetID"] = sourceTargetDf["target"].apply(
        lambda x: labelList.index(x)
    )

    # creating data for the sankey diagram
    data = dict(
        type="sankey",
        node=dict(
            pad=15,
            thickness=20,
            line=dict(color="black", width=0.5),
            label=labelList,
            color=colorList,
        ),
        link=dict(
            source=sourceTargetDf["sourceID"],
            target=sourceTargetDf["targetID"],
            value=sourceTargetDf["count"],
        ),
    )

    # override gray link colors with 'source' colors
    opacity = 0.4
    # change 'magenta' to its 'rgba' value to add opacity
    data["node"]["color"] = [
        "rgba(255,0,255, 0.8)" if color == "magenta" else color
        for color in data["node"]["color"]
    ]
    data["link"]["color"] = [
        data["node"]["color"][src].replace("0.8", str(opacity))
        for src in data["link"]["source"]
    ]

    fig = go.Figure(
        data=[
            go.Sankey(
                # Define nodes
                node=dict(
                    pad=15,
                    thickness=15,
                    line=dict(color="black", width=0.5),
                    label=data["node"]["label"],
                    color=data["node"]["color"],
                ),
                # Add links
                link=dict(
                    source=data["link"]["source"],
                    target=data["link"]["target"],
                    value=data["link"]["value"],
                    color=data["link"]["color"],
                ),
            )
        ]
    )

    fig.update_layout(title_text=title, font_size=10)

    return fig



import os
import requests
from typing import Dict


def get_rocrate_metadata_gh(sample_id: str) -> Dict:
    """
    Retrieves RO-Crate metadata from a GitHub repository.

    Args:
        sample_id (str): The ID of the sample.

    Returns:
        Dict: The metadata in JSON format.
    """
    url = f"https://api.github.com/repos/emo-bon/analysis-results-cluster-01-crate/contents/{sample_id}-ro-crate/ro-crate-metadata.json"
    req = requests.get(
        url,
        headers={
            "accept": "application/vnd.github.v3.raw",
        },
    )
    print("ro-crate-metadata.json request status", req.status_code)
    return req.json()


def get_rocrate_data(metadata_json: Dict, data_id: str):
    """
    Retrieves RO-Crate data file based on metadata.

    Args:
        metadata_json (Dict): The metadata in JSON format.
        data_id (str): The ID of the data file.
    Returns:
        str: The content of the data file.
    """
    raise NotImplementedError


def extract_data_by_name(metadata: Dict, name: str) -> Dict:
    """
    Extracts data from metadata based on the name.
    Args:
        metadata (Dict): The metadata in JSON format.
        name (str): The name of the data to extract.
    Returns:
        Dict: The extracted data.
    """
    for d in metadata["@graph"]:
        if "name" in d.keys() and d["name"] == name:
            return d
    return None


def extract_all_datafiles(metadata: Dict) -> list:
    """
    Extracts all data files from the metadata.
    Args:
        metadata (Dict): The metadata in JSON format.
    Returns:
        List[Dict]: A list of dictionaries containing data file information.
    """
    datafiles = []
    for d in metadata["@graph"]:
        if "name" in d.keys() and d["@type"] == "File":
            data_unit = {}
            data_unit["name"] = d["name"]
            try:
                # in MB
                data_unit["sizeMB"] = int(int(d["contentSize"]) / 1e6)
            except KeyError:
                data_unit["sizeMB"] = "unknown"

            try:
                data_unit["downloadUrl"] = d["downloadUrl"]
            except KeyError:
                data_unit["downloadUrl"] = "unknown"
            datafiles.append(data_unit)

    return datafiles



import os
import sys
import psutil
import logging
from IPython import get_ipython


#####################
# Environment setup #
#####################
# TODO: there needs to be a yaml file to set up a folder structure, hardcoding here is not good :)
# Question: Or should this be part of the momics package?
def init_setup():
    """
    Initializes the setup environment.

    This function checks if the current environment is IPython (such as Google Colab).
    If it is, it runs the setup for IPython environments. Otherwise, it runs the setup
    for local environments.
    """
    # First solve if IPython
    if is_ipython():
        ## For running at GColab, the easiest is to clone and then pip install some deps
        setup_ipython()


def install_colab_packages():
    try:
        os.system("pip install panel hvplot")
        print(f"panel and hvplot installed")
    except OSError as e:
        print(f"An error occurred while installing panel and hvplot: {e}")


def setup_ipython():
    """
    Setup the IPython environment.

    This function installs the momics package and other dependencies for the IPython environment.
    """
    if "google.colab" in str(get_ipython()):
        print("Google Colab")

        # Install ngrok for hosting the dashboard
        try:
            os.system("pip install pyngrok --quiet")
            print("ngrok installed")
        except OSError as e:
            print(f"An error occurred while installing ngrok: {e}")

        # Install the momics package
        install_colab_packages()


def is_ipython():
    # This is for the case when the script is run from the Jupyter notebook
    if "ipykernel" not in sys.modules:
        print("Not an IPython setup")
        return False

    return True


def get_notebook_environment():
    """
    Determine if the notebook is running in VS Code or JupyterLab.

    Returns:
        str: The environment in which the notebook is running ('vscode', 'jupyter:binder', 'jupyter:local' or 'unknown').
    """
    # Check for VS Code environment variable
    if "VSCODE_PID" in os.environ:
        return "vscode"

    elif "JPY_SESSION_NAME" in os.environ:
        if psutil.users() == []:
            print("Binder")
            return "jupyter:binder"
        else:
            print("Local Jupyter")
            return "jupyter:local"
    else:
        return "unknown"


###########
# logging #
###########
FORMAT = "%(levelname)s | %(name)s | %(message)s"  # for logger


def reconfig_logger(format=FORMAT, level=logging.INFO):
    """(Re-)configure logging"""
    logging.basicConfig(format=format, level=level, force=True)

    # removing tarnado access logs
    hn = logging.NullHandler()
    logging.getLogger("tornado.access").addHandler(hn)
    logging.getLogger("tornado.access").propagate = False

    logging.info("Logging.basicConfig completed successfully")


#####################
# Memory management #
#####################
def memory_load():
    """
    Get the memory usage of the current process.

    Returns:
        tuple: A tuple containing:
            - used_gb (float): The amount of memory currently used by the process in gigabytes.
            - total_gb (float): The total amount of memory available in gigabytes.
    """
    used_gb, total_gb = (
        psutil.virtual_memory()[3] / 1000000000,
        psutil.virtual_memory()[0] / 1000000000,
    )
    return used_gb, total_gb


def memory_usage():
    """
    Get the memory usage of the current process.

    Returns:
        list: A list of tuples containing the names of the objects in the current environment
            and their corresponding sizes in bytes.
    """
    # These are the usual ipython objects, including this one you are creating
    ipython_vars = ["In", "Out", "exit", "quit", "get_ipython", "ipython_vars"]

    # Get a sorted list of the objects and their sizes
    mem_list = sorted(
        [
            (x, sys.getsizeof(globals().get(x)))
            for x in dir()
            if not x.startswith("_") and x not in sys.modules and x not in ipython_vars
        ],
        key=lambda x: x[1],
        reverse=True,
    )

    return mem_list
